/*
 *  John W. Linville, linville@lvl7.com
 *  Copyright (C) 2003 LVL7 Systems, Inc.  All rights reserved.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/config.h>

#ifdef CONFIG_PCI

#include <linux/types.h>
#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/init.h>

#include <asm/pci_channel.h>
#include <asm/paccess.h>
#include <asm/delay.h>

#define PCI_ACCESS_READ  0
#define PCI_ACCESS_WRITE 1

#if 1 /*JWL*/
#undef DEBUG
#else
#define DEBUG
#endif

#ifdef DEBUG
#define DBG(x...) printk(KERN_ERR x)
#else
#define DBG(x...)
#endif

u32 bcm47xx_idmask = 0x0000ffff;

#define MEM_BASE    0x08000000
#define MEM_SIZE    0x03ffffff
#define IO_BASE     0x0c000000
#define IO_SIZE     0x03ffffff

static struct resource pci_io_resource = {
	"PCI I/O", 
	IO_BASE,
	IO_BASE + IO_SIZE,
	IORESOURCE_IO};

static struct resource pci_mem_resource = {
	"PCI Memory", 
	MEM_BASE,
	MEM_BASE + MEM_SIZE,
	IORESOURCE_MEM};

extern struct pci_ops bcm47xx_pci_ops;

/*
 * The mips_pci_channels array has all descriptors for all
 * pci bus controllers. Usually on most boards there's only
 * one pci controller to worry about.
 *
 * Note that the '0x08' and '0xff' below indicate the first
 * and last "devfn" to scan.  You can use these variables
 * to limit the scan (e.g. we skip the PCI bridge).
 */
struct pci_channel mips_pci_channels[] = {
	{ &bcm47xx_pci_ops, &pci_io_resource, &pci_mem_resource, 0x08, 0xff },
	{ (struct pci_ops *)NULL, (struct resource *)NULL,
	  (struct resource *)NULL, (int)NULL, (int)NULL}
};

/*
 * Typically there is one core config routine which the pci_ops
 * functions call.
 */
static int
bcm47xx_config_access(unsigned char access_type, struct pci_dev *dev,
                      unsigned char where, u32 *data)
{
  u32 addr;

  /*
   * IMPORTANT
   * If a pci config cycle fails, it's *very* important
   * that if the cycle requested is READ, you set *data
   * to 0xffffffff. The pci_auto code does not check the
   * return value of the pci_ops functions. It expects that
   * if a pci config cycle read fails, the data returned
   * will be 0xffffffff.
   */

  if (PCI_SLOT(dev->devfn) != 16) {
    if (access_type == PCI_ACCESS_READ) {
      *data = 0xffffffff;
    }
    return(-1);
  }
  
  if (dev->bus->number == 0) {
    /* bgh - this maps to config space for device 16 (idsel 15) */
    addr = 0xac008000;
    addr |= (where & 0xfc);

    if (access_type == PCI_ACCESS_READ) {
      udelay(10000);
      *data = readl(addr);
      udelay(10000);
    } else {
      udelay(10000);
      writel(*data, addr);
      udelay(10000);
    }
    return(0);
  
  } else {
    if (access_type == PCI_ACCESS_READ) {
      *data = 0xffffffff;
    }
    return(-1);
  }
}


static int
read_config_byte (struct pci_dev *dev, int where, u8 *val)
{
	u32 data = 0;

	if (bcm47xx_config_access(PCI_ACCESS_READ, dev, where, &data)) {
		*val = 0xff;
		return -1;
	}

	*val = (data >> ((where & 3) << 3)) & 0xff;
        DBG("cfg read byte: bus %d dev_fn %x where %x: val %x\n", 
                dev->bus->number, dev->devfn, where, *val);

	return PCIBIOS_SUCCESSFUL;
}


static int
read_config_word (struct pci_dev *dev, int where, u16 *val)
{
	u32 data = 0;

	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;

	if (bcm47xx_config_access(PCI_ACCESS_READ, dev, where, &data)) {
		*val = 0xffff;
		return -1;
	}

	*val = (data >> ((where & 3) << 3)) & 0xffff;
        DBG("cfg read word: bus %d dev_fn %x where %x: val %x\n", 
                dev->bus->number, dev->devfn, where, *val);

	return PCIBIOS_SUCCESSFUL;
}

static int
read_config_dword (struct pci_dev *dev, int where, u32 *val)
{
	u32 data = 0;

	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	
	if (bcm47xx_config_access(PCI_ACCESS_READ, dev, where, &data)) {
		*val = 0xffffffff;
		return -1;
	}

	*val = data;
        DBG("cfg read dword: bus %d dev_fn %x where %x: val %x\n", 
                dev->bus->number, dev->devfn, where, *val);

	return PCIBIOS_SUCCESSFUL;
}


static int
write_config_byte (struct pci_dev *dev, int where, u8 val)
{
	u32 data = 0;
       
	if (bcm47xx_config_access(PCI_ACCESS_READ, dev, where, &data))
		return -1;

	data = (data & ~(0xff << ((where & 3) << 3))) |
	       (val << ((where & 3) << 3));
        DBG("cfg write byte: bus %d dev_fn %x where %x: val %x\n", 
                dev->bus->number, dev->devfn, where, val);

	if (bcm47xx_config_access(PCI_ACCESS_WRITE, dev, where, &data))
		return -1;

	return PCIBIOS_SUCCESSFUL;
}

static int
write_config_word (struct pci_dev *dev, int where, u16 val)
{
        u32 data = 0;

	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;
       
        if (bcm47xx_config_access(PCI_ACCESS_READ, dev, where, &data))
	       return -1;

	data = (data & ~(0xffff << ((where & 3) << 3))) | 
	       (val << ((where & 3) << 3));
        DBG("cfg write word: bus %d dev_fn %x where %x: val %x\n", 
                dev->bus->number, dev->devfn, where, val);

	if (bcm47xx_config_access(PCI_ACCESS_WRITE, dev, where, &data))
	       return -1;


	return PCIBIOS_SUCCESSFUL;
}

static int
write_config_dword(struct pci_dev *dev, int where, u32 val)
{
	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;

	if (bcm47xx_config_access(PCI_ACCESS_WRITE, dev, where, &val))
	       return -1;
        DBG("cfg write dword: bus %d dev_fn %x where %x: val %x\n", 
                dev->bus->number, dev->devfn, where, val);

	return PCIBIOS_SUCCESSFUL;
}

struct pci_ops bcm47xx_pci_ops = {
	read_config_byte,
        read_config_word,
	read_config_dword,
	write_config_byte,
	write_config_word,
	write_config_dword
};

#endif /* CONFIG_PCI */
