 /*
  * i2c-algo-85xx.c i2c driver algorithms for the MPC85xx I2C Interface
  * Derived from Wolfgang Denk's MPC8240 I2C driver
  * 
  * (C) Copyright 2002,2003 Motorola Inc
  * Xianghua Xiao (x.xiao@motorola.com)
  *
  * (C) Copyright 2000 DENX Software Engineering
  * Wolfgang Denk, wd@denx.de
  * All rights reserved.
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */

#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-85xx.h>

#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif
#include <linux/module.h>

#define DEB(x)      if (i2c_debug >= 1) x
#define DEB2(x)     if (i2c_debug >= 2) x
#define DEB3(x)     if (i2c_debug >= 3) x 
#define DEBPROTO(x) if (i2c_debug >= 9) x

#define DEF_TIMEOUT 50

static int i2c_debug  = 0;
static int mpc85xx_scan = 0;
static wait_queue_head_t i2c_wait;
static int mpc85xx_i2c_poll = 0;
static int mpc85xx_i2c_pending = 0;

/* Read from I2C registers */
static u_char mpc85xx_i2c_read(volatile u_char * addr)
{
	u_char val;

	val = in_8(addr);
	asm volatile ("msync");
	return val;
}


/* Write to I2C registers, mask is used to keep reserved bit safe(except I2CDR) */
static void mpc85xx_i2c_write(volatile u_char * addr, u_char val, u_char mask)
{
	u_char tmp;
	
	if (! mask) {
		out_8(addr, val);
	} else {
		tmp = mpc85xx_i2c_read(addr);
		asm volatile ("msync");
		out_8(addr, (tmp & ~mask) | (val & mask));
	}

	asm volatile ("msync");
}


/* I2C interrupt handler */
static void
mpc85xx_i2c_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	volatile ccsr_i2c_t *i2c = (ccsr_i2c_t *)dev_id;

	DEB2(printk("mpc85xx_i2c_interrupt(dev_id=%p)\n", dev_id));

	mpc85xx_i2c_write(&i2c->i2csr, ~MPC85xx_I2CSR_MIF, MPC85xx_I2CSR_MIF);

	mpc85xx_i2c_pending = 1;

	wake_up_interruptible(&i2c_wait);
}


/* Generate START on I2C bus */
static void i2c_start(struct i2c_algo_85xx_data *adap) 
{
	volatile ccsr_i2c_t *i2c = (ccsr_i2c_t *)adap->i2c;

	DEBPROTO(printk("Start I2C\n"));
	mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_MSTA, MPC85xx_I2CCR_MSTA);
}


/* Generate STOP on I2C bus */
static void i2c_stop(struct i2c_algo_85xx_data *adap) 
{
	volatile ccsr_i2c_t *i2c = (ccsr_i2c_t *)adap->i2c;

	DEBPROTO(printk("Stop I2C\n"));
	mpc85xx_i2c_write(&i2c->i2ccr, 0, MPC85xx_I2CCR_MSTA);
}


/* Check I2C bus busy condition */
static int wait_for_bb(struct i2c_algo_85xx_data *adap)
{
	volatile ccsr_i2c_t *i2c = (ccsr_i2c_t *)adap->i2c;
	int timeout = DEF_TIMEOUT;
	u_char status;

	status = mpc85xx_i2c_read(&i2c->i2csr);
#ifndef STUB_I2C
	while (timeout-- && (status & MPC85xx_I2CSR_MBB)) {
		udelay(1000);
		status = status = mpc85xx_i2c_read(&i2c->i2csr);
	}
#endif
	if (timeout <= 0) {
		DEB2(printk("Timeout waiting for Bus Busy\n"));
	}

	return (timeout <= 0);
}


static inline void mpc85xx_sleep(unsigned long timeout)
{
	schedule_timeout( timeout * HZ);
}


/* Check three conditions based on I2CSR[MIF]:
   0: No interrupt pending
   1: One byte transferd, or
      I2CADR matches slave-receiver mode's calling address, or
      Arbitration is lost
*/
static inline int wait_for_pin(struct i2c_algo_85xx_data *adap, u_char *status)
{
	int timeout = DEF_TIMEOUT;
	int res     = -1;
	volatile ccsr_i2c_t *i2c = (ccsr_i2c_t *)adap->i2c;
	unsigned long flags;

	if (! mpc85xx_i2c_poll)	{
		DEB2(printk("wait for pin: sleep on interrupt\n"));
		
		save_flags(flags); 
		cli();
		if(mpc85xx_i2c_pending == 0){
			interruptible_sleep_on_timeout(&i2c_wait,1*HZ);
		} else {
			mpc85xx_i2c_pending = 0;
		}

		restore_flags(flags);

		if(signal_pending(current))
			return -ERESTARTSYS;

		*status = mpc85xx_i2c_read(&i2c->i2csr);
		res = 0;
		goto Done;
	} else {
		*status = mpc85xx_i2c_read(&i2c->i2csr);
#ifndef STUB_I2C
		while (timeout-- && !(*status & MPC85xx_I2CSR_MIF)) {
			udelay(1000);
			*status = mpc85xx_i2c_read(&i2c->i2csr);
		}
#endif

		if (!(*status & MPC85xx_I2CSR_MIF)) {
			goto Done; /* res = -1 */
		} else {
			mpc85xx_i2c_write(&i2c->i2csr, ~MPC85xx_I2CSR_MIF, MPC85xx_I2CSR_MIF);
			DEB2(printk("wait for pin: poll done\n"));
			res = 0;
		}
	}

Done:
	return res;
}


/* I2C Initialization */
static void
mpc85xx_i2c_init(struct i2c_algo_85xx_data *adap)
{
	volatile ccsr_i2c_t	*i2c = adap->i2c;

	DEB2(printk("mpc85xx_i2c_init() - I2C registers at %p\n", i2c));

	/* Select an arbitrary address.  Just make sure it is unique. */
	mpc85xx_i2c_write(&i2c->i2cadr, I2C_MPC85xx_ADDR, MPC85xx_I2CADR_MASK);

	/* Make clock run maximum slow.	*/
	mpc85xx_i2c_write(&i2c->i2cfdr, 0x3f, MPC85xx_I2CFDR_MASK);

	/* Set the default digital filter sampling rate. */
	mpc85xx_i2c_write(&i2c->i2cdfsrr,0x00, MPC85xx_I2CDFSRR);

	/* Enable master/transmit. */
	mpc85xx_i2c_write(&i2c->i2ccr, ~MPC85xx_I2CCR_MSTA, MPC85xx_I2CCR_MSTA);
	mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_MTX, MPC85xx_I2CCR_MTX);

	/* Clear interrupt bit */
	mpc85xx_i2c_write(&i2c->i2csr, ~MPC85xx_I2CSR_MIF, MPC85xx_I2CSR_MIF);

	/* init the wait queue */
	init_waitqueue_head(&i2c_wait);

	/* Install interrupt handler. */
	DEB2(printk ("%s[%d] Install ISR for IRQ %d\n",__func__,__LINE__, MPC85xx_IRQ_IIC));

	if (request_irq(MPC85xx_IRQ_IIC, mpc85xx_i2c_interrupt, 0, "i2c",(void *)i2c)) {
		DEB(printk(KERN_WARNING "Failed to install an inetrrupt handler "
		       " for MPC85xx I2C contorller IRQ%d\n", MPC85xx_IRQ_IIC));
		mpc85xx_i2c_poll = 1;
	} else {
		if(mpc85xx_i2c_poll == 0) {
			mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_MIEN,
		        	        MPC85xx_I2CCR_MIEN);
		} else {
			mpc85xx_i2c_write(&i2c->i2ccr, ~MPC85xx_I2CCR_MIEN,
		        	        MPC85xx_I2CCR_MIEN);
		}	
	}

	/* Enable I2C Interface */
	mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_MEN, MPC85xx_I2CCR_MEN);
}


static int try_address(struct i2c_algo_85xx_data *adap,
		       unsigned char addr, int retries)
{
	volatile ccsr_i2c_t	*i2c = adap->i2c;
	int i, ret = -1;
	u_char status;

	for (i = 0; i < retries; i++) {
		i2c_start(adap);
		mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_MTX, MPC85xx_I2CCR_MTX);
		mpc85xx_i2c_write(&i2c->i2cdr, addr, 0);
		
		if (wait_for_pin(adap, &status) >= 0) {
			if (!(status & MPC85xx_I2CSR_RXAK)) { 
				DEB2(printk("(0x%02x)", addr >> 1));
				i2c_stop(adap);
				ret = 0;
				goto Done;
			} else {
				DEB2(printk("."));
			}
		}
		i2c_stop(adap);
		udelay(500);
	}
Done:
	DEB2(if (i < retries)
	      printk("i2c-algo-85xx.o: needed %d retries for %d\n",i,addr));
	return ret;
}


/* Send bytes to I2C slave */
static int mpc85xx_sendbytes(struct i2c_adapter *i2c_adap, const char *buf,
                         int count)
{
	struct i2c_algo_85xx_data *adap = i2c_adap->algo_data;
	volatile ccsr_i2c_t	  *i2c  = adap->i2c;
	int wrcount, timeout;
	unsigned char status;

  	mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_MTX, MPC85xx_I2CCR_MTX);
  
	for (wrcount = 0; wrcount < count; ++wrcount) {
		DEB2(printk("i2c-algo-85xx.o: %s i2c_write: writing %dth byte = 0x%x\n",
		      i2c_adap->name, wrcount+1,buf[wrcount]&0xff));
		mpc85xx_i2c_write(&i2c->i2cdr, buf[wrcount], 0);
		timeout = wait_for_pin(adap, &status);
		if (timeout) {
			i2c_stop(adap);
			DEB2(printk("i2c-algo-85xx.o: %s i2c_write: "
			       "error - timeout.\n", i2c_adap->name));
			return -EREMOTEIO; /* got a better one ?? */
		}
#ifndef STUB_I2C
		if (status & MPC85xx_I2CSR_RXAK) {
			i2c_stop(adap);
			DEB2(printk("i2c-algo-85xx.o: %s i2c_write: "
			       "error - no ack.\n", i2c_adap->name));
			return -EREMOTEIO;
		}
#endif
	}
	return (wrcount);
}

/*
 * Description: Called by the upper layers to do the grunt work for
 * a master read transaction.
 * Note: The No-ACK should be generated by write I2CCR[TXAK] before I read 
 * 'last-to-next' byte. Here I read the very 'last' one per current 
 * Puff model design.Currently I read 16 bytes once a time.The Figure 11-11
 * in Dracom UM Rev2.0 suggested 1 byte read needs to be handled differently...
 *
 */
static int mpc85xx_readbytes(struct i2c_adapter *i2c_adap, char *buf, int count)
{
	int i, timeout;
	unsigned char status;
	struct i2c_algo_85xx_data *adap = i2c_adap->algo_data;
	volatile ccsr_i2c_t	  *i2c  = adap->i2c;

	/* change to master receive mode */    
	mpc85xx_i2c_write(&i2c->i2ccr, 0, MPC85xx_I2CCR_MTX);

	if(count == 1) mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_TXAK, MPC85xx_I2CCR_TXAK);
	/* dummy read */
	mpc85xx_i2c_read(&i2c->i2cdr);

	for (i = 0; i < count;i++) {
		timeout = wait_for_pin(adap, &status);
		if (timeout) {
			i2c_stop(adap);
			DEB2(printk("i2c-algo-85xx.o: i2c_read timed out.\n"));
			return -EREMOTEIO;
		}

		if(i == count-2) {
			/* generate txack on next to last byte */
			mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_TXAK, MPC85xx_I2CCR_TXAK);
		}
		
		if(i == count-1) {
			/* generate stop on last byte */
			i2c_stop(adap);
		}
		
		buf[i] = mpc85xx_i2c_read(&i2c->i2cdr);
	}
	
	return count;
}


/*
 * Description: Whenever we initiate a transaction, the first byte clocked
 * onto the bus after the start condition is the address (7 bit) of the
 * device we want to talk to.  This function manipulates the address specified
 * so that it makes sense to the hardware when written to the IIC peripheral.
 *
 * Note: 10 bit addresses are not supported in this driver, although they are
 * supported by the hardware.  This functionality needs to be implemented.
 */
static inline void mpc85xx_doAddress(struct i2c_algo_85xx_data *adap,
                                struct i2c_msg *msg, int retries) 
{
	volatile ccsr_i2c_t *i2c = (ccsr_i2c_t *)adap->i2c;
	unsigned short flags = msg->flags;
	unsigned char addr;

	addr = ( msg->addr << 1 );

	if(msg->addr == I2C_MPC85xx_ADDR) {
		DEB(printk("Warning: client's address is the same as master\n"));
	}

	if (flags & I2C_M_RD ) {
		addr |= 1;
	}

	if (flags & I2C_M_REV_DIR_ADDR ) {
			addr ^= 1;
	}

	mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_MTX, MPC85xx_I2CCR_MTX);

	if(mpc85xx_i2c_poll == 0) {
		mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_MIEN,
	        	        MPC85xx_I2CCR_MIEN);
	} else {
		mpc85xx_i2c_write(&i2c->i2ccr, ~MPC85xx_I2CCR_MIEN,
	        	        MPC85xx_I2CCR_MIEN);
	}

	mpc85xx_i2c_write(&i2c->i2cdr, addr, 0);
	DEB2(printk("I2C slave address 0x%x sent\n",addr));
}

/*
 * Description:  This function implements combined transactions.  Combined
 * transactions consist of combinations of reading and writing blocks of data.
 * Each transfer (i.e. a read or a write) is separated by a repeated start
 * condition.
 * The combination could be R-R or W-W or R-W or W-R. only the final message(R or W)
 * does not need a Repeat-START.
 *
 * Note: always prepare and fill I2CDR before set I2CCR[RSTA] per Puff design.
 * Note: always do a write first then RESTART for a read(EEPROM specific).
 */
static int
mpc85xx_combined_transaction(struct i2c_adapter *i2c_adap, struct i2c_msg msgs[],
			 int num)
{
	int i,timeout;
	unsigned char status;
	struct i2c_msg *pmsg;
	int ret;
	int restart_flag = 0;
	struct i2c_algo_85xx_data *adap = i2c_adap->algo_data;
	volatile ccsr_i2c_t *i2c = (ccsr_i2c_t *)adap->i2c;

	DEB2(printk("Beginning combined transaction\n"));
	for (i = 0; i < num; i++) {
		pmsg = &msgs[i];

		/* if a RESTART happened, we need send an address first */
		if(restart_flag == 1) {
			DEB3(printk("i2c-algo-85xx.o: total Msg: %d Msg %d, addr=0x%x, flags=0x%x, "
				    "len=%d\n",num,i, msgs[i].addr, msgs[i].flags, msgs[i].len);)

			mpc85xx_i2c_write(&i2c->i2csr, ~MPC85xx_I2CSR_MIF, MPC85xx_I2CSR_MIF);

			status = mpc85xx_i2c_read(&i2c->i2csr);
			DEB2(printk("before RESTART address cycle, i2csr = 0x%x\n",status));

			mpc85xx_doAddress(adap, pmsg, i2c_adap->retries);

			timeout = wait_for_pin(adap, &status);

			if (timeout) {
				i2c_stop(adap);
				DEB2(printk("i2c-algo-85xx.o: Timeout waiting "
					    "for i2csr[MIF] not set in mpc85xx_combined_transaction\n");)
			}
#ifndef STUB_I2C
			/* make sure we received the ACK from slave right after address cycle */
			if (status & MPC85xx_I2CSR_RXAK) {
				i2c_stop(adap);
				DEB2(printk("i2c-algo-85xx.o: status=%d   "
					    "No RXAK in mpc85xx_combined_transaction\n",status);)
			}
			restart_flag = 0;
		}
#endif

		if (pmsg->flags & I2C_M_RD) {
			ret = mpc85xx_readbytes(i2c_adap, pmsg->buf, pmsg->len);

			if (ret != pmsg->len) {
				DEB2(printk("i2c-algo-85xx.o: fail: "
					    "only read %d bytes.\n", ret));
				return i;
			} else {
				DEB2(printk
				     ("i2c-algo-85xx.o: read %d bytes.\n",
				      ret));
			}

			if( i < num - 1 ) {
				DEB2(printk("Repeat START at read\n"));
				mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_RSTA, MPC85xx_I2CCR_RSTA);
				restart_flag = 1;
			} else {
				DEB2(printk("last in read/write---no need for Repeat START\n"));
			}


		} else if (!(pmsg->flags & I2C_M_RD)) {

			ret = mpc85xx_sendbytes(i2c_adap, pmsg->buf, pmsg->len);

			if (ret != pmsg->len) {
				DEB2(printk("i2c-algo-85xx.o: fail: "
					    "only wrote %d bytes.\n", ret));
				return i;
			} else {
				DEB2(printk
				     ("i2c-algo-85xx.o: wrote %d bytes.\n",
				      ret));
			}

			if( i < num - 1 ) {
				DEB2(printk("Repeat START at write\n"));
				mpc85xx_i2c_write(&i2c->i2ccr, MPC85xx_I2CCR_RSTA, MPC85xx_I2CCR_RSTA);
				restart_flag = 1;
			} else {
				DEB2(printk("last in read/write---no need for Repeat START\n"));
			}
		}
	}

	return num;
}



/* This is the entry to initiate I2C bus read/write to/from I2C slaves*/
static int mpc85xx_xfer(struct i2c_adapter *i2c_adap,
		    struct i2c_msg msgs[], 
		    int num)
{
	struct i2c_algo_85xx_data * adap            = i2c_adap->algo_data;
	struct i2c_msg            * pmsg;
	int                         i, ret;
	int                         timeout;
	unsigned char		    status;
	volatile ccsr_i2c_t *i2c = (ccsr_i2c_t *)adap->i2c;

	i=0;
	pmsg = &msgs[i];

	DEB3(printk("i2c-algo-85xx.o: total Msg: %d Msg %d, addr=0x%x, flags=0x%x, "
	            "len=%d\n",
	            num,i, msgs[i].addr, msgs[i].flags, msgs[i].len);)

	/* Three steps to generate 'START': */
	/* 1:Check for bus busy.*/
	timeout = wait_for_bb(adap);
	if (timeout) {
		DEB2(printk("i2c-algo-85xx.o: "
			    "Timeout waiting for BB in mpc85xx_xfer\n");)
		ret = -EIO;
		goto Done;
	}

	/* 2: clear interrupt and set i2ccr[MSTA] to select master mode */
	mpc85xx_i2c_write(&i2c->i2csr, ~MPC85xx_I2CSR_MIF, MPC85xx_I2CSR_MIF);
	i2c_start(adap);
	status = mpc85xx_i2c_read(&i2c->i2csr);
	DEB2(printk("before transfer, i2csr = 0x%x\n",status));
  
	/* 3: write slave address to i2cdr[0-6], set i2ccr[MTX] as direction */
	mpc85xx_doAddress(adap, pmsg, i2c_adap->retries);

	/* The above three steps complete the address cycle */

	timeout = wait_for_pin(adap, &status);

	if (timeout) {
		i2c_stop(adap);
		DEB2(printk("i2c-algo-85xx.o: Timeout waiting "
		            "for PIN(1) in mpc85xx_xfer\n");)
		ret = EREMOTEIO;
		goto Done;
	}
    
#ifndef STUB_I2C
	/* make sure we received the ACK from slave right after address cycle */
	if (status & MPC85xx_I2CSR_RXAK) {
		i2c_stop(adap);
		DEB2(printk("i2c-algo-85xx.o: status=%d   "
		            "No RXAK in mpc85xx_xfer\n",status);)
		ret = -EREMOTEIO;
		goto Done;
	}
#endif

	if ( num > 1 ) {
		DEB2(printk("i2c-algo-85xx.o: do combined transaction\n"));
		ret = mpc85xx_combined_transaction(i2c_adap,msgs,num);
	} else if ( ( num==1 )&&( pmsg->flags & I2C_M_RD )) {
		DEB2(printk("i2c-algo-85xx.o: do read transaction(non-combine)\n"));
		ret = mpc85xx_readbytes(i2c_adap, pmsg->buf, pmsg->len);
        	if (ret != pmsg->len) {
			DEB2(printk("i2c-algo-85xx.o: fail: "
			            "only read %d bytes.\n", ret));
			ret = (ret < 0) ? ret : -EREMOTEIO;
			goto Done;
		} else {
			DEB2(printk("i2c-algo-85xx.o: read %d bytes.\n",ret));
		}
	} else if ( ( num==1) && ( !(pmsg->flags & I2C_M_RD))) {
		DEB2(printk("i2c-algo-85xx.o: do write transaction(non-combine)\n"));
		ret = mpc85xx_sendbytes(i2c_adap, pmsg->buf, pmsg->len);
		if (ret != pmsg->len) {
			DEB2(printk("i2c-algo-85xx.o: fail: "
			            "only wrote %d bytes.\n", ret));
			ret = (ret < 0) ? ret : -EREMOTEIO;
			goto Done;
		} else {
			DEB2(printk("i2c-algo-85xx.o: wrote %d "
			            "bytes.\n", ret));
		}
	} else {
		DEB2(printk("no transfer after address cycle, such as SMBUS_QUICK mode\n"));
	}

	ret = num;

Done:
	i2c_stop(adap);
	return ret;
}


static int algo_control(struct i2c_adapter *adapter, 
	unsigned int cmd, unsigned long arg)
{
	return 0;
}

static u32 mpc85xx_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_SMBUS_EMUL; 
}

/* -----exported algorithm data: -------------------------------------	*/

static struct i2c_algorithm mpc85xx_algo = {
	"MPC85xx algorithm",
	I2C_ALGO_MPC85xx,
	mpc85xx_xfer,
	NULL,
	NULL,				/* slave_xmit		*/
	NULL,				/* slave_recv		*/
	algo_control,			/* ioctl		*/
	mpc85xx_func,			/* functionality	*/
};

/* 
 * registering functions to load algorithms at runtime 
 */
int i2c_mpc85xx_add_bus(struct i2c_adapter *adap)
{
	int i;
	struct i2c_algo_85xx_data *mpc85xx_adap = adap->algo_data;

	DEB2(printk("i2c-algo-85xx.o: hw routines for %s registered.\n",
	            adap->name));

	adap->id |= mpc85xx_algo.id;
	adap->algo = &mpc85xx_algo;

	adap->timeout = DEF_TIMEOUT;
	adap->retries = 3;

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif

	i2c_add_adapter(adap);
	mpc85xx_i2c_init(mpc85xx_adap);

	if (mpc85xx_scan) {
		DEB(printk(KERN_INFO " i2c-algo-85xx.o: scanning bus %s.\n",
		       adap->name));
		for (i = 0x00; i < 0xff; i+=2) {
			try_address(mpc85xx_adap, i, 1);
			udelay(500);
		}
		DEB(printk("\n"));
	}
	return 0;
}


int i2c_mpc85xx_del_bus(struct i2c_adapter *adap)
{
	int res;
	if ((res = i2c_del_adapter(adap)) < 0)
		return res;
	DEB2(printk("i2c-algo-85xx.o: adapter unregistered: %s\n",adap->name));

#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

int __init i2c_algo_mpc85xx_init (void)
{
	printk("i2c-algo-85xx.o: i2c MPC85xx algorithm module\n");
	return 0;
}

EXPORT_SYMBOL(i2c_mpc85xx_add_bus);
EXPORT_SYMBOL(i2c_mpc85xx_del_bus);

#ifdef MODULE
MODULE_AUTHOR("Xianghua Xiao <x.xiao@motorola.com>");
MODULE_DESCRIPTION("I2C-Bus MPC85xx algorithm");

MODULE_PARM(mpc85xx_scan, "i");
MODULE_PARM(i2c_debug,"i");

MODULE_PARM_DESC(mpc85xx_scan, "Scan for active chips on the bus");
MODULE_PARM_DESC(i2c_debug,
        "debug level - 0 off; 1 normal; 2,3 more verbose; 9 mpc85xx-protocol");


int init_module(void) 
{
	return i2c_algo_mpc85xx_init();
}

void cleanup_module(void) 
{
	free_irq(MPC85xx_IRQ_IIC,0);
}
#endif

