/*
 * arch/ppc/kernel/mv64360_common.c
 *
 * Common routines for the Marvell/Galileo MV64360 (Discovery II) host bridge,
 * interrupt controller, memory controller, serial controller, enet controller,
 * etc.
 *
 * Written by Rabeeh Khoury - rabeeh@galileo.co.il
 *
 * Based on work done by Mark A. Greer <mgreer@mvista.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/slab.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/machdep.h>
#include <asm/pci-bridge.h>
#include <asm/mv64360.h>
#include <asm/delay.h>


u32	mv64360_base; /* Virtual base address of MV64360's regs */
u32	mv64360_revision; /* Revision of the chip */
u8	mv64360_pci_exclude_bridge = TRUE;

static int num_hoses;

/*
 *****************************************************************************
 *
 *	Bridge Initialization Routines
 *
 *****************************************************************************
 */
static void mv64360_check_errata(struct pci_controller *hose_a,
				 struct pci_controller *hose_b);

/*
 * Typical '_find_bridges()' routine for boards with a MV64360 bridge.
 */
int __init
mv64360_find_bridges(u32 phys_base_addr, mv64360_bridge_info_t *info,
	int ((*map_irq)(struct pci_dev *, unsigned char, unsigned char)),
	int enumerate)
{
	struct pci_controller		*hose_a, *hose_b;
	u32				io_base_a, io_base_b;
	int				rc;
	u32				val;
	u8 				dev;

	if ( ppc_md.progress )
		ppc_md.progress("mv64360_find_bridges: enter", 0);

	mv64360_base = (u32)ioremap(phys_base_addr,MV64360_INTERNAL_SPACE_SIZE);

	if (info->pci_0_mem_size == 0 && info->pci_0_io_size == 0)
		hose_a = NULL;
	else {
		hose_a = pcibios_alloc_controller();
		if (!hose_a) {
			if (ppc_md.progress)
				ppc_md.progress("ev64260_setup_bridge: hose_a failed", 0);

			return -1;
		}
		num_hoses++;
	}
	if (info->pci_1_mem_size == 0 && info->pci_1_io_size == 0)
		hose_b = NULL;
	else {
		hose_b = pcibios_alloc_controller();
		if (!hose_b) {
			if (ppc_md.progress)
				ppc_md.progress("ev64260_setup_bridge: hose_b failed", 0);

			return -1;
		}
		num_hoses++;
	}
	if (num_hoses == 0)
		printk("mv64360_find_bridges: No PCI hoses\n");

	info->hose_a = hose_a;
	info->hose_b = hose_b;
	
	/* Set PCI0 device number to 0 */
	if (hose_a) {
		val = MV_REG_READ (MV64360_PCI_0_P2P_CONFIG);
		dev = ((val >> 24) & 0xff);
		if (dev) {
			val = mv64360_hw_read_config_dword(
					0, 0, dev, 0, 0x64, 0);
			val = val & 0xffffff00;
			mv64360_hw_write_config_dword(
					0, 0, dev, 0, 0x64, 0, val);
		}
	}
	
	/* Set PCI1 device number to 0*/
	if (hose_b) {
		val = MV_REG_READ (MV64360_PCI_1_P2P_CONFIG);
		dev = ((val >> 24) & 0xff);
		if (dev) {
			val = mv64360_hw_read_config_dword(
					1, 0, dev, 0, 0x64, 0);
			val = val & 0xffffff00;
			mv64360_hw_write_config_dword (
					1, 0, dev, 0, 0x64, 0, val);
		}
	}

	/* Ends up mapping PCI Config addr/data pairs twice */
	if (hose_a) {
		setup_indirect_pci(hose_a,
			phys_base_addr + MV64360_PCI_0_CONFIG_ADDR,
			phys_base_addr + MV64360_PCI_0_CONFIG_DATA_VIRTUAL_REG);
	}

	if (hose_b) {
		setup_indirect_pci(hose_b,
			phys_base_addr + MV64360_PCI_1_CONFIG_ADDR,
			phys_base_addr + MV64360_PCI_1_CONFIG_DATA_VIRTUAL_REG);
	}


	if ((rc = mv64360_bridge_init(info)) != 0) {
		if (ppc_md.progress)
			ppc_md.progress("mv64360_bridge_init: failed ", 0);

		if (hose_a) {
			iounmap((void *)hose_a->cfg_addr);
			iounmap((void *)hose_a->cfg_data);
		}
		if (hose_b) {
			iounmap((void *)hose_b->cfg_addr);
			iounmap((void *)hose_b->cfg_data);
		}
		return rc;
	}

	/* ioremap PCI I/O regions */
	if (hose_b) {
		io_base_b = (u32)ioremap(
				info->pci_1_io_start_proc, info->pci_1_io_size);
	}
	if (hose_a) {
		io_base_a = (u32)ioremap(
				info->pci_0_io_start_proc, info->pci_0_io_size);
		isa_io_base = io_base_a;
		hose_a->first_busno = 0;
		hose_a->last_busno = 0xff;

		/* Set PCI0 bus number to 0 */
		if (MV_REG_READ (MV64360_PCI_0_MODE) & (BIT4 | BIT5)) {
			/* Relevant when working in PCI-X mode */
			val = mv64360_hw_read_config_dword(0, 0, 0, 0, 0x64, 0);
			val = val & 0xffff00ff;
			mv64360_hw_write_config_dword (0, 0, 0, 0, 0x64, 0, val);
		} else {
			/* Relevant when working in PCI rev 2.2 */
			val = MV_REG_READ (MV64360_PCI_0_P2P_CONFIG);
			val = val & 0xff00ffff;
			MV_REG_WRITE (MV64360_PCI_0_P2P_CONFIG, val);
		}

		pci_init_resource(&hose_a->io_resource,
			0,	/* really: io_base_a - isa_io_base */
			info->pci_0_io_size - 1,
			IORESOURCE_IO,
			"host bridge PCI bus 0");
		hose_a->io_space.start = info->pci_0_io_start_pci;
		hose_a->io_space.end   = info->pci_0_io_start_pci +
						info->pci_0_io_size - 1;
		hose_a->io_base_virt = (void *)isa_io_base;

		pci_init_resource(&hose_a->mem_resources[0],
			  info->pci_0_mem_start_proc,
			  info->pci_0_mem_start_proc + info->pci_0_mem_size - 1,
			  IORESOURCE_MEM,
			  "host bridge PCI bus 0");
		hose_a->mem_space.start = info->pci_0_mem_start_pci_lo;
		hose_a->mem_space.end   = info->pci_0_mem_start_pci_lo +
						info->pci_0_mem_size - 1;
		hose_a->pci_mem_offset = (info->pci_0_mem_start_proc -
						info->pci_0_mem_start_pci_lo);

		if (enumerate) {
			hose_a->last_busno =
				pciauto_bus_scan(hose_a, hose_a->first_busno);
		}
	}

	if (hose_b) {
		if (hose_a)
			hose_b->first_busno = hose_a->last_busno + 1;
		else {
			hose_b->first_busno = 0;
			isa_io_base = io_base_b;
		}
		hose_b->last_busno  = 0xff;

		/* Set PCI1 bus number to hose_a->last_busno + 1 */
		if (MV_REG_READ (MV64360_PCI_1_MODE) & (BIT4 | BIT5)) {
			/* Relevant when working in PCI-X mode */
			val = mv64360_hw_read_config_dword(1, 0, 0, 0, 0x64, 0);
			val = (val & 0xffff00ff) |
			    		(((hose_b->first_busno) & 0xff) << 8);
			mv64360_hw_write_config_dword(1, 0, 0, 0, 0x64, 0, val);
		} else {
			/* Relevant when working in PCI rev 2.2 */
			val = MV_REG_READ (MV64360_PCI_1_P2P_CONFIG);
			val = (val & 0xff00ffff) | (((hose_b->first_busno) & 0xff) << 16);
			MV_REG_WRITE (MV64360_PCI_1_P2P_CONFIG, val);
		}

		pci_init_resource(&hose_b->io_resource,
			  io_base_b - isa_io_base,
			  io_base_b - isa_io_base + info->pci_1_io_size - 1,
			  IORESOURCE_IO,
			  "host bridge PCI bus 1");
		hose_b->io_space.start = info->pci_1_io_start_pci;
		hose_b->io_space.end   = info->pci_1_io_start_pci +
						info->pci_1_io_size - 1;
		hose_b->io_base_virt = (void *)isa_io_base;

		pci_init_resource(&hose_b->mem_resources[0],
			  info->pci_1_mem_start_proc,
			  info->pci_1_mem_start_proc + info->pci_1_mem_size - 1,
			  IORESOURCE_MEM,
			  "host bridge PCI bus 1");
		hose_b->mem_space.start = info->pci_1_mem_start_pci_lo;
		hose_b->mem_space.end   = info->pci_1_mem_start_pci_lo +
					info->pci_1_mem_size - 1;
		hose_b->pci_mem_offset = (info->pci_1_mem_start_proc -
					info->pci_1_mem_start_pci_lo);

		if (enumerate) {
			hose_b->last_busno =
				pciauto_bus_scan(hose_b, hose_b->first_busno);
		}
	}

	ppc_md.pci_exclude_device = mv64360_pci_exclude_device;
	ppc_md.pci_swizzle        = common_swizzle;
	ppc_md.pci_map_irq        = map_irq;

	mv64360_pci_disable_all_windows(info->hose_a, info->hose_b);

	return 0;
} /* mv64360_find_bridges() */

#ifdef CONFIG_NOT_COHERENT_CACHE
#define ACCESS_CONTROL_FEATURES MV64360_PCI_ACC_CNTL_NONE_CACHE_COHERENT
#else
#define ACCESS_CONTROL_FEATURES MV64360_PCI_ACC_CNTL_CACHE_COHERENT
#endif

/*
 * mv64360_bridge_init()
 *
 * Perform bridge initialization for a "typical" setup for a PPC system.
 */
int __init
mv64360_bridge_init(mv64360_bridge_info_t *info)
{
	int	window;
	u32	u32_val;
	int	rc = 0;
	u8	save_exclude;
	
	if ( ppc_md.progress )
		ppc_md.progress("mv64360_bridge_init: enter", 0);

	/*
	 * Turn off timer/counters.  Not turning off watchdog timer because
	 * can't read its reg on the MV64360 so don't know if we'll be enabling
	 * or disabling.
	 */
	MV_RESET_REG_BITS (MV64360_TIMER_COUNTER_0_3_CONTROL,
			   (BIT0 | BIT8 | BIT16 | BIT24));
	
	save_exclude = mv64360_pci_exclude_bridge;
	mv64360_pci_exclude_bridge = FALSE;

	if (info->hose_a) {
		/* PCI 0 Timeout and Retry limits */
		u32_val = MV_REG_READ(MV64360_PCI_0_RETRY);
		u32_val |= 0x00ffffff;
		MV_REG_WRITE(MV64360_PCI_0_RETRY, u32_val);
		
		/* Set class code to indicate host bridge */
		u32_val = mv64360_hw_read_config_dword (0, 0, 0, 0, PCI_CLASS_REVISION, 0);
		u32_val &= 0x000000ff;
		mv64360_revision = u32_val;
		u32_val |= (PCI_CLASS_BRIDGE_HOST << 16);
		mv64360_hw_write_config_dword (0, 0, 0, 0, PCI_CLASS_REVISION, 0, u32_val);

		/* Enable MV64360 to be PCI master & respond to PCI MEM cycles */
		u32_val = mv64360_hw_read_config_dword (0, 0, 0, 0, PCI_COMMAND, 0);
		u32_val |= (PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY | PCI_COMMAND_IO);
		mv64360_hw_write_config_dword (0, 0, 0, 0, PCI_COMMAND, 0, u32_val);

		/* Set cacheline size and latency timer (to 128)
		 * clear BIST
		 * set multi func
		 */
		u32_val = mv64360_hw_read_config_dword (0, 0, 0, 0, PCI_CACHE_LINE_SIZE, 0);
		u32_val &= 0xffffff00;
		u32_val |= L1_CACHE_LINE_SIZE / 4;
#define CHANGE_MAX_LATENCY
#ifdef CHANGE_MAX_LATENCY
		u32_val &= 0xffff00ff;
		u32_val |= 0x80 << 8;
#endif
		mv64360_hw_write_config_dword (0, 0, 0, 0, PCI_CACHE_LINE_SIZE, 0, u32_val);
	}

	if (info->hose_b) {
		/* PCI 1 Timeout and Retry limits */
		u32_val = MV_REG_READ(MV64360_PCI_1_RETRY);
		u32_val |= 0x00ffffff;
		MV_REG_WRITE(MV64360_PCI_1_RETRY, u32_val);

		/* Set class code to indicate host bridge */
		u32_val = mv64360_hw_read_config_dword (1, 0, 0, 0, PCI_CLASS_REVISION, 0);
		u32_val &= 0x000000ff;
		u32_val |= (PCI_CLASS_BRIDGE_HOST << 16);
		mv64360_hw_write_config_dword (1, 0, 0, 0, PCI_CLASS_REVISION, 0, u32_val);

		/* Enable MV64360 to be PCI master & respond to PCI MEM cycles */
		u32_val = mv64360_hw_read_config_dword (1, 0, 0, 0, PCI_COMMAND, 0);
		u32_val |= (PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY | PCI_COMMAND_IO);
		mv64360_hw_write_config_dword (1, 0, 0, 0, PCI_COMMAND, 0, u32_val);

		/* Set cacheline size and latency timer (to 128)
		 * clear BIST
		 * set multi func
		 */
		u32_val = mv64360_hw_read_config_dword (1, 0, 0, 0, PCI_CACHE_LINE_SIZE, 0);
		u32_val &= 0xffffff00;
		u32_val |= L1_CACHE_LINE_SIZE / 4;
#ifdef CHANGE_MAX_LATENCY
		u32_val &= 0xffff00ff;
		u32_val |= 0x80 << 8;
#endif
		mv64360_hw_write_config_dword (1, 0, 0, 0, PCI_CACHE_LINE_SIZE, 0, u32_val);
	}
	mv64360_pci_exclude_bridge = save_exclude;


	/*
	 * Disable all CPU windows on the bridge except for DDR CS.
	 */
	if ( ppc_md.progress )
		ppc_md.progress("mv64360_bridge_init: disable all CPU decodings except memory access", 0);
	
	mv64360_cpu_disable_all_windows();

	if (info->hose_a) {
		/*
		 * Set up CPU->PCI mappings (so CPU can get at PCI dev
		 * regs/mem).  Will only use one of the four CPU->PCI
		 * MEM windows on each bus.
		 */
		mv64360_cpu_set_pci_io_window(0,
					      info->pci_0_io_start_proc,
					      info->pci_0_io_start_pci,
					      info->pci_0_io_size,
					      info->pci_0_io_swap);

		mv64360_cpu_set_pci_mem_window(0,
					       0,
					       info->pci_0_mem_start_proc,
					       info->pci_0_mem_start_pci_hi,
					       info->pci_0_mem_start_pci_lo,
					       info->pci_0_mem_size,
					       info->pci_0_mem_swap);

		/*
		 * For virt_to_bus & bus_to_virt to work correctly, this
		 * mapping must be the same on both PCI buses.
		 */
		mv64360_pci_slave_cs_set_window(info->hose_a,
						0,
						0x00000000,
						0x00000000,
						0x00000000,
						0x00000000,
						info->mem_size);

		/* Disable all the access control windows */
		for (window=0; window<MV64360_PCI_ACC_CNTL_WINDOWS; window++)
			mv64360_pci_acc_cntl_set_window(0, window, 0, 0, 0, 0);

		mv64360_pci_acc_cntl_set_window(0,
						0,
						0x00000000,
						0x00000000,
						info->mem_size,
						ACCESS_CONTROL_FEATURES);
	}

	if (info->hose_b) {
		/*
		 * Set up CPU->PCI mappings (so CPU can get at PCI dev
		 * regs/mem).  Will only use one of the four CPU->PCI
		 * MEM windows on each bus.
		 */
		mv64360_cpu_set_pci_io_window(1,
					      info->pci_1_io_start_proc,
					      info->pci_1_io_start_pci,
					      info->pci_1_io_size,
					      info->pci_1_io_swap);

		mv64360_cpu_set_pci_mem_window(1,
					       0,
					       info->pci_1_mem_start_proc,
					       info->pci_1_mem_start_pci_hi,
					       info->pci_1_mem_start_pci_lo,
					       info->pci_1_mem_size,
					       info->pci_1_mem_swap);
	
		/*
		 * For virt_to_bus & bus_to_virt to work correctly, this
		 * mapping must be the same on both PCI buses.
		 */
		mv64360_pci_slave_cs_set_window(info->hose_b,
						0,
						0x00000000,
						0x00000000,
						0x00000000,
						0x00000000,
						info->mem_size);

		/* Disable all the access control windows */
		for (window=0; window<MV64360_PCI_ACC_CNTL_WINDOWS; window++)
			mv64360_pci_acc_cntl_set_window(1, window, 0, 0, 0, 0);

		mv64360_pci_acc_cntl_set_window(1,
						0,
						0x00000000,
						0x00000000,
						info->mem_size,
						ACCESS_CONTROL_FEATURES);
	}

	pci_dram_offset = 0; /* System mem at same addr on PCI & cpu bus */

	mv64360_check_errata(info->hose_a, info->hose_b);

	return rc;
} /* mv64360_bridge_init() */

/*
 * mv64360_check_errata()
 *
 * Apply applicable errata and restrictions from 0.5 of the
 * Errata and Restrictions document from Marvell/Galileo.
 */
static void __init
mv64360_check_errata(struct pci_controller *hose_a,
		     struct pci_controller *hose_b)
{
} /* mv64360_check_errata() */


/*
 *****************************************************************************
 *
 *	General Window Setting Routines
 *
 *****************************************************************************
 */

static int
mv64360_set_32bit_window(u32 base_addr,
			 u32 size,
			 u32 access_flags,
			 u32 bot_reg,
			 u32 top_reg)
{
    u32	enablebit = 0xffff;
    unsigned int new_base_addr, new_size, temp, shift;

    switch(bot_reg)
      {
      case MV64360_CS_0_BASE_ADDR:
	enablebit = BIT0;
	break;
      case MV64360_CS_1_BASE_ADDR:
	enablebit = BIT1;
	break;
      case MV64360_CS_2_BASE_ADDR:
	enablebit = BIT2;
	break;
      case MV64360_CS_3_BASE_ADDR:
	enablebit = BIT3;
	break;
      case MV64360_DEV_CS0_BASE_ADDR:
	enablebit = BIT4;
	break;
      case MV64360_DEV_CS1_BASE_ADDR:
	enablebit = BIT5;
	break;
      case MV64360_DEV_CS2_BASE_ADDR:
	enablebit = BIT6;
	break;
      case MV64360_DEV_CS3_BASE_ADDR:
	enablebit = BIT7;
	break;
      case MV64360_BOOTCS_BASE_ADDR:
	enablebit = BIT8;
	break;
      case MV64360_PCI_0_IO_BASE_ADDR:
	enablebit = BIT9;
	break;
      case MV64360_PCI_0_MEMORY0_BASE_ADDR:
	enablebit = BIT10;
	break;
      case MV64360_PCI_0_MEMORY1_BASE_ADDR:
	enablebit = BIT11;
	break;
      case MV64360_PCI_0_MEMORY2_BASE_ADDR:
	enablebit = BIT12;
	break;
      case MV64360_PCI_0_MEMORY3_BASE_ADDR:
	enablebit = BIT13;
	break;
      case MV64360_PCI_1_IO_BASE_ADDR:
	enablebit = BIT14;
	break;
      case MV64360_PCI_1_MEMORY0_BASE_ADDR:
	enablebit = BIT15;
	break;
      case MV64360_PCI_1_MEMORY1_BASE_ADDR:
	enablebit = BIT16;
	break;
      case MV64360_PCI_1_MEMORY2_BASE_ADDR:
	enablebit = BIT17;
	break;
      case MV64360_PCI_1_MEMORY3_BASE_ADDR:
	enablebit = BIT18;
	break;
      case MV64360_INTEGRATED_SRAM_BASE_ADDR:
	enablebit = BIT19;
	break;
      case MV64360_INTERNAL_SPACE_BASE_ADDR:
	enablebit = BIT20;
	break;
      case MV64360_CPU_PROTECT_WINDOW_0_BASE_ADDR:
	if (size == 0) MV_REG_WRITE (MV64360_CPU_PROTECT_WINDOW_0_BASE_ADDR, 0);
	return 0;
      case MV64360_CPU_PROTECT_WINDOW_1_BASE_ADDR:
	if (size == 0) MV_REG_WRITE (MV64360_CPU_PROTECT_WINDOW_1_BASE_ADDR, 0);
	return 0;
      case MV64360_CPU_PROTECT_WINDOW_2_BASE_ADDR:
	if (size == 0) MV_REG_WRITE (MV64360_CPU_PROTECT_WINDOW_2_BASE_ADDR, 0);
	return 0;
      case MV64360_CPU_PROTECT_WINDOW_3_BASE_ADDR:
	if (size == 0) MV_REG_WRITE (MV64360_CPU_PROTECT_WINDOW_3_BASE_ADDR, 0);
	return 0;
      }

    /* If setting integrated SRAM or internal registers, don't check size */
    if ((enablebit == BIT19) && (enablebit == BIT20)) {
	    base_addr = base_addr >> 16;
	    MV_REG_WRITE (bot_reg, base_addr | access_flags);
	    return 0;
    }
    /* If the size is 0, the window will be disabled */
    if (size == 0) {
	    if (enablebit != 0xffff) {
		    MV_SET_REG_BITS (MV64360_BASE_ADDR_ENABLE, enablebit);
		    return 0;
	    }
	    /* Although window is disabled, map window to address 0 */
	    MV_REG_WRITE(bot_reg,0xb000);
	    /* top_reg can be 0 in case of mapping internal SRAM */
	    if (top_reg != 0)
		    MV_REG_WRITE(top_reg,0);
	    wmb();
	    return 0;

    }
    /* The base address must be aligned to the size */
    if ((base_addr % size) != 0) {
	    return -1;
    }
    if (size >= _64K) {
	    new_base_addr = base_addr >> 16;
	    new_size = size >> 16;
	    /* Checking that the size is a sequence of '1' followed by a
	       sequence of '0' starting from LSB to MSB. */
	    temp = new_size - 1;
	    for(shift = 0; shift < 16; shift++) {
		    temp = temp >> shift;
		    if((temp & 0x1) == 0) {
			    /* Either we got to the last '1' */
			    /* or the size is not valid       */
			    if(temp > 0x0)
				    return -1;
			    else
				    break;
		    }
	    }
	    /* Write the new values */
	    /* bot_reg must be set all the time */
	    MV_REG_WRITE(bot_reg,new_base_addr);
	    /* top_reg can be 0 in case of mapping internal SRAM */
	    if (top_reg != 0)
		    MV_REG_WRITE(top_reg,new_size - 1);
	    /* Enable the window */
	    if (enablebit != 0xffff)
		    MV_RESET_REG_BITS (MV64360_BASE_ADDR_ENABLE, enablebit);
	    /* Memory sync barrier */
	    wmb();
	    return 0;
    }
    /* Given size is smaller than 64KByte */
    return -1;
}


/*
 *****************************************************************************
 *
 *	CPU Configuration Routines
 *
 *****************************************************************************
 */

int
mv64360_cpu_cs_set_window(u32 window,
			   u32 base_addr,
			   u32 size)
{
	static u32
	cpu_cs_windows[MV64360_CPU_CS_DECODE_WINDOWS][2] = {
		{ MV64360_CS_0_BASE_ADDR, MV64360_CS_0_SIZE },
		{ MV64360_CS_1_BASE_ADDR, MV64360_CS_1_SIZE },
		{ MV64360_CS_2_BASE_ADDR, MV64360_CS_2_SIZE },
		{ MV64360_CS_3_BASE_ADDR, MV64360_CS_3_SIZE },
	}; /* cpu_cs_windows[][] */
	int	rc = -1;

	if (window < MV64360_CPU_CS_DECODE_WINDOWS) {
		rc = mv64360_set_32bit_window(base_addr,
					      size,
					      0,
					      cpu_cs_windows[window][0],
					      cpu_cs_windows[window][1]);
	}

	return rc;
} /* mv64360_cpu_cs_set_window() */

int
mv64360_cpu_dev_cs_set_window(u32 window,
			  u32 base_addr,
			  u32 size)
{
	static u32
	cpu_dev_cs_windows[MV64360_CPU_DEV_CS_DECODE_WINDOWS][2] = {
		{ MV64360_DEV_CS0_BASE_ADDR, MV64360_DEV_CS0_SIZE },
		{ MV64360_DEV_CS1_BASE_ADDR, MV64360_DEV_CS1_SIZE },
		{ MV64360_DEV_CS2_BASE_ADDR, MV64360_DEV_CS2_SIZE },
		{ MV64360_DEV_CS3_BASE_ADDR, MV64360_DEV_CS3_SIZE },
	}; /* cpu_dev_cs_windows[][] */
	int	rc = -1;

	if (window < MV64360_CPU_DEV_CS_DECODE_WINDOWS) {
		rc = mv64360_set_32bit_window(base_addr,
					      size,
					      0,
					      cpu_dev_cs_windows[window][0],
					      cpu_dev_cs_windows[window][1]);
	}

	return rc;
} /* mv64360_cpu_dev_cs_set_window() */

int
mv64360_cpu_boot_set_window(u32 base_addr,
			    u32 size)
{
	int	rc;

	rc = mv64360_set_32bit_window(base_addr,
				      size,
				      0,
				      MV64360_BOOTCS_BASE_ADDR,
				      MV64360_BOOTCS_SIZE);

	return rc;
} /* mv64360_cpu_boot_set_window() */

int
mv64360_cpu_internal_sram_set_window(u32 base_addr)
{
	int	rc = -1;
	rc = mv64360_set_32bit_window(base_addr,
				      _256K,
				      0,
				      MV64360_INTEGRATED_SRAM_BASE_ADDR,
				      0);

	return rc;
} /* mv64360_cpu_internal_sram_set_window() */


/*
 * mv64360_cpu_set_pci_io_window()
 *
 * Set up a CPU window into PCI I/O or MEM space.
 * Always do Read/Modify/Write to window regs.
 */
static int
mv64360_cpu_pci_set_window(u32 cpu_base_addr,
			   u32 pci_base_addr,
			   u32 size,
			   u32 access_flags,
			   u32 bot_reg,
			   u32 top_reg,
			   u32 remap_reg)
{
	u32	val;
	int	rc;

	if ((rc = mv64360_set_32bit_window(cpu_base_addr,
					   size,
					   access_flags,
					   bot_reg,
					   top_reg)) == 0) {

		if (size == 0) {
			MV_REG_WRITE(remap_reg, 0xffffffff);
		} else {
			MV_REG_WRITE(remap_reg, pci_base_addr >> 16);
		}
		val = MV_REG_READ(bot_reg); /* Flush FIFO */
	}

	return rc;
} /* mv64360_cpu_pci_set_window() */


/*
 * mv64360_cpu_set_pci_io_window()
 *
 * Set up a CPU window into PCI I/O space.
 * Always do Read/Modify/Write to window regs.
 */
int
mv64360_cpu_set_pci_io_window(u32 pci_bus,
			      u32 cpu_base_addr,
			      u32 pci_base_addr,
			      u32 size,
			      u32 swap)
{
	/* 2 PCI buses with 1 I/O window each (from CPU point of view) */
	static u32
	cpu_pci_io_windows[MV64360_PCI_BUSES][4] = {
		{ MV64360_PCI_0_IO_BASE_ADDR,
		  MV64360_PCI_0_IO_SIZE,
		  MV64360_PCI_0_IO_ADDR_REMAP },

		{ MV64360_PCI_1_IO_BASE_ADDR,
		  MV64360_PCI_1_IO_SIZE,
		  MV64360_PCI_1_IO_ADDR_REMAP },
	}; /* cpu_pci_io_windows[][] */
	int	rc = -1;

	if (pci_bus < MV64360_PCI_BUSES) {
		rc =  mv64360_cpu_pci_set_window(cpu_base_addr,
					  pci_base_addr,
					  size,
					  swap,
					  cpu_pci_io_windows[pci_bus][0],
					  cpu_pci_io_windows[pci_bus][1],
					  cpu_pci_io_windows[pci_bus][2]);
	}

	return rc;
} /* mv64360_cpu_set_pci_io_window() */

/*
 * mv64360_cpu_set_pci_mem_window()
 *
 * Set up a CPU window into PCI MEM space (4 PCI MEM windows per PCI bus).
 * Always do Read/Modify/Write to window regs.
 */
int
mv64360_cpu_set_pci_mem_window(u32 pci_bus,
			       u32 window,
			       u32 cpu_base_addr,
			       u32 pci_base_addr_hi,
			       u32 pci_base_addr_lo,
			       u32 size,
			       u32 swap_64bit)
{
	/* 2 PCI buses with 4 memory windows each (from CPU point of view) */
	static u32
	cpu_pci_mem_windows[MV64360_PCI_BUSES][MV64360_PCI_MEM_WINDOWS_PER_BUS][4] = {
		{ /* PCI 0 */
			{ MV64360_PCI_0_MEMORY0_BASE_ADDR,
			  MV64360_PCI_0_MEMORY0_SIZE,
			  MV64360_PCI_0_MEMORY0_LOW_ADDR_REMAP,
			  MV64360_PCI_0_MEMORY0_HIGH_ADDR_REMAP },

			{ MV64360_PCI_0_MEMORY1_BASE_ADDR,
			  MV64360_PCI_0_MEMORY1_SIZE,
			  MV64360_PCI_0_MEMORY1_LOW_ADDR_REMAP,
			  MV64360_PCI_0_MEMORY1_HIGH_ADDR_REMAP },

			{ MV64360_PCI_0_MEMORY2_BASE_ADDR,
			  MV64360_PCI_0_MEMORY2_SIZE,
			  MV64360_PCI_0_MEMORY2_LOW_ADDR_REMAP,
			  MV64360_PCI_0_MEMORY2_HIGH_ADDR_REMAP },

			{ MV64360_PCI_0_MEMORY3_BASE_ADDR,
			  MV64360_PCI_0_MEMORY3_SIZE,
			  MV64360_PCI_0_MEMORY3_LOW_ADDR_REMAP,
			  MV64360_PCI_0_MEMORY3_HIGH_ADDR_REMAP },
		},

		{ /* PCI 1 */
			{ MV64360_PCI_1_MEMORY0_BASE_ADDR,
			  MV64360_PCI_1_MEMORY0_SIZE,
			  MV64360_PCI_1_MEMORY0_LOW_ADDR_REMAP,
			  MV64360_PCI_1_MEMORY0_HIGH_ADDR_REMAP },

			{ MV64360_PCI_1_MEMORY1_BASE_ADDR,
			  MV64360_PCI_1_MEMORY1_SIZE,
			  MV64360_PCI_1_MEMORY1_LOW_ADDR_REMAP,
			  MV64360_PCI_1_MEMORY1_HIGH_ADDR_REMAP },

			{ MV64360_PCI_1_MEMORY2_BASE_ADDR,
			  MV64360_PCI_1_MEMORY2_SIZE,
			  MV64360_PCI_1_MEMORY2_LOW_ADDR_REMAP,
			  MV64360_PCI_1_MEMORY2_HIGH_ADDR_REMAP },

			{ MV64360_PCI_1_MEMORY3_BASE_ADDR,
			  MV64360_PCI_1_MEMORY3_SIZE,
			  MV64360_PCI_1_MEMORY3_LOW_ADDR_REMAP,
			  MV64360_PCI_1_MEMORY3_HIGH_ADDR_REMAP }
		}
	}; /* cpu_pci_mem_windows[][][] */
	u32		remap_reg, remap;
	int		rc = -1;

	if ((pci_bus < MV64360_PCI_BUSES) &&
	    (window < MV64360_PCI_MEM_WINDOWS_PER_BUS)) {

		if (mv64360_cpu_pci_set_window(
			cpu_base_addr,
			pci_base_addr_lo,
			size,
			swap_64bit,
			cpu_pci_mem_windows[pci_bus][window][0],
			cpu_pci_mem_windows[pci_bus][window][1],
			cpu_pci_mem_windows[pci_bus][window][2]) == 0) {

			remap_reg = cpu_pci_mem_windows[pci_bus][window][3];
			MV_REG_WRITE(remap_reg, pci_base_addr_hi);

			remap = MV_REG_READ(remap_reg); /* Flush FIFO */

			rc = 0;
		}
	}

	return rc;
} /* mv64360_cpu_set_pci_mem_window() */

int
mv64360_cpu_prot_set_window(u32 window,
			    u32 base_addr,
			    u32 size,
			    u32 access_bits)
{
	static u32
	cpu_prot_windows[MV64360_CPU_PROT_WINDOWS][2] = {
		{ MV64360_CPU_PROTECT_WINDOW_0_BASE_ADDR, MV64360_CPU_PROTECT_WINDOW_0_SIZE },
		{ MV64360_CPU_PROTECT_WINDOW_1_BASE_ADDR, MV64360_CPU_PROTECT_WINDOW_1_SIZE },
		{ MV64360_CPU_PROTECT_WINDOW_2_BASE_ADDR, MV64360_CPU_PROTECT_WINDOW_2_SIZE },
		{ MV64360_CPU_PROTECT_WINDOW_3_BASE_ADDR, MV64360_CPU_PROTECT_WINDOW_3_SIZE }
	}; /* cpu_prot_windows[][] */
	int	rc = -1;

	if (window < MV64360_CPU_PROT_WINDOWS) {
		rc = mv64360_set_32bit_window(base_addr,
					      size,
					      access_bits,
					      cpu_prot_windows[window][0],
					      cpu_prot_windows[window][1]);
	}

	return rc;
} /* mv64360_cpu_prot_set_window() */


void
mv64360_pci_disable_all_windows(struct pci_controller *hose_a, struct pci_controller *hose_b)
{
	int	window;

	/* Disable DDR CS mapping */
	for (window = 0 ; window < MV64360_CPU_CS_DECODE_WINDOWS; window++) {
		if (hose_a)
			mv64360_pci_slave_cs_set_window (
					hose_a, window, 0, 0, 0, 0, 0);
		if (hose_b)
			mv64360_pci_slave_cs_set_window (
					hose_b, window, 0, 0, 0, 0, 0);
	}

	for (window = 0 ; window < MV64360_CPU_DEV_CS_DECODE_WINDOWS; window++) {
		if (hose_a)
			mv64360_pci_slave_dev_cs_set_window (
					hose_a, window, 0, 0, 0, 0);
		if (hose_b)
			mv64360_pci_slave_dev_cs_set_window (
					hose_b, window, 0, 0, 0, 0);
	}
	if (hose_a) {
		mv64360_pci_slave_boot_set_window (hose_a, 0, 0, 0, 0);
		mv64360_pci_slave_internal_sram_set_window (hose_a, 0, 0, 0, 0);
	}
	if (hose_b) {
		mv64360_pci_slave_boot_set_window (hose_b, 0, 0, 0, 0);
		mv64360_pci_slave_internal_sram_set_window (hose_b, 0, 0, 0, 0);
	}
	for (window = 0 ; window < MV64360_PCI_P2P_MEM_WINDOWS; window++) {
		if (hose_a)
			mv64360_pci_slave_p2p_mem_set_window (
					hose_a ,window, 0, 0, 0, 0, 0);
		if (hose_b)
			mv64360_pci_slave_p2p_mem_set_window (
					hose_b ,window, 0, 0, 0, 0, 0);
	}
	if (hose_a)
		mv64360_pci_slave_p2p_io_set_window (hose_a, 0, 0, 0);
	if (hose_b)
		mv64360_pci_slave_p2p_io_set_window (hose_b, 0, 0, 0);

	/* Close all bar enable bits */
	if (hose_a)
		mv64360_pci_bar_enable (0, 0);
	if (hose_b)
		mv64360_pci_bar_enable (1, 0);
}




	
void
mv64360_cpu_disable_all_windows(void)
{
	int	pci_bus, window;

	/* Don't disable CS windows b/c we need to access system memory */
	for (window=0; window<MV64360_CPU_DEV_CS_DECODE_WINDOWS; window++) {
		mv64360_cpu_dev_cs_set_window(window, 0, 0);
	}

	mv64360_cpu_boot_set_window(0, 0);

	for (pci_bus=0; pci_bus<num_hoses; pci_bus++) {
		mv64360_cpu_set_pci_io_window(pci_bus, 0, 0, 0, 0);

		for (window=0;window<MV64360_PCI_MEM_WINDOWS_PER_BUS;window++) {
			mv64360_cpu_set_pci_mem_window(pci_bus,
						       window,
						       0, 0, 0, 0, 0);
		}
	}

	for (window=0; window<MV64360_CPU_PROT_WINDOWS; window++) {
		mv64360_cpu_prot_set_window(window, 0, 0, 0);
	}

#if 0 /* Obsolete */
	for (window=0; window<MV64360_CPU_SNOOP_WINDOWS; window++) {
		mv64360_cpu_snoop_set_window(window, 0, 0, 0);
	}
#endif

	return;
} /* mv64360_cpu_disable_all_windows() */


/*
 *****************************************************************************
 *
 *	PCI Slave Window Configuration Routines
 *
 *****************************************************************************
 */

int
mv64360_pci_bar_enable(u32 pci_bus,
		       u32 enable_bits)
{
	u32	reg, val;
	int	rc = -1;

	if (pci_bus < MV64360_PCI_BUSES) {
		reg = (pci_bus == 0) ? MV64360_PCI_0_BASE_ADDR_REG_ENABLE :
				       MV64360_PCI_1_BASE_ADDR_REG_ENABLE;


		/* Note: '0' enables, '1' disables */
		val = MV_REG_READ(reg);
		val |= 0xffffffff;	/* Disable everything by default */
		val &= ~enable_bits;
		MV_REG_WRITE(reg, val);

		MV_REG_READ(reg); /* Flush FIFO */

		rc = 0;
	}

	return rc;
} /* mv64360_pci_bar_enable() */

static int
mv64360_pci_slave_set_window(struct pci_controller *hose,
			     u32 pci_base_addr,
			     u32 pci_base_addr_hi,
			     u32 cpu_base_addr,
			     u32 cpu_base_addr_hi,
			     u32 bar_size,
			     u32 pci_cfg_fcn,
			     u32 pci_cfg_hdr_offset,
			     u32 bar_size_reg,
			     u32 remap_reg,
			     u32 remap_reg_hi)
{
	u32	val, enablebit = 0xffff;
	int	devfn;
	u8	save_exclude;
	u8      pci_bus = (hose->first_busno == 0) ? 0 : 1;

	pci_base_addr &= 0xfffff000;
	cpu_base_addr &= 0xfffff000;
	bar_size &= 0xfffff000;

	if ((pci_cfg_fcn == 0) && (pci_cfg_hdr_offset == 0x20)) {
		enablebit = BIT9;
	}
	
	switch(remap_reg)
		{
		case MV64360_PCI_0_CS_0_BASE_ADDR_REMAP:
		case MV64360_PCI_1_CS_0_BASE_ADDR_REMAP:
			enablebit = BIT0;
			break;
		case MV64360_PCI_0_CS_1_BASE_ADDR_REMAP:
		case MV64360_PCI_1_CS_1_BASE_ADDR_REMAP:
			enablebit = BIT1;
			break;
		case MV64360_PCI_0_CS_2_BASE_ADDR_REMAP:
		case MV64360_PCI_1_CS_2_BASE_ADDR_REMAP:
			enablebit = BIT2;
			break;
		case MV64360_PCI_0_CS_3_BASE_ADDR_REMAP:
		case MV64360_PCI_1_CS_3_BASE_ADDR_REMAP:
			enablebit = BIT3;
			break;
		case MV64360_PCI_0_DEVCS_0_BASE_ADDR_REMAP:
		case MV64360_PCI_1_DEVCS_0_BASE_ADDR_REMAP:
			enablebit = BIT4;
			break;
		case MV64360_PCI_0_DEVCS_1_BASE_ADDR_REMAP:
		case MV64360_PCI_1_DEVCS_1_BASE_ADDR_REMAP:
			enablebit = BIT5;
			break;
		case MV64360_PCI_0_DEVCS_2_BASE_ADDR_REMAP:
		case MV64360_PCI_1_DEVCS_2_BASE_ADDR_REMAP:
			enablebit = BIT6;
			break;
		case MV64360_PCI_0_DEVCS_3_BASE_ADDR_REMAP:
		case MV64360_PCI_1_DEVCS_3_BASE_ADDR_REMAP:
			enablebit = BIT7;
			break;
		case MV64360_PCI_0_DEVCS_BOOTCS_BASE_ADDR_REMAP:
		case MV64360_PCI_1_DEVCS_BOOTCS_BASE_ADDR_REMAP:
			enablebit = BIT8;
			break;
			/* Missing here internal registers map */
		case MV64360_PCI_0_P2P_MEM0_BASE_ADDR_REMAP_LOW:
		case MV64360_PCI_1_P2P_MEM0_BASE_ADDR_REMAP_LOW:
			enablebit = BIT11;
			break;
		case MV64360_PCI_0_P2P_MEM1_BASE_ADDR_REMAP_LOW:
		case MV64360_PCI_1_P2P_MEM1_BASE_ADDR_REMAP_LOW:
			enablebit = BIT12;
			break;
		case MV64360_PCI_0_P2P_I_O_BASE_ADDR_REMAP:
		case MV64360_PCI_1_P2P_I_O_BASE_ADDR_REMAP:
			enablebit = BIT13;
			break;
		case MV64360_PCI_0_CPU_BASE_ADDR_REMAP_LOW:
		case MV64360_PCI_1_CPU_BASE_ADDR_REMAP_LOW:
			enablebit = BIT14;
			break;
		case MV64360_PCI_0_INTEGRATED_SRAM_BASE_ADDR_REMAP:
		case MV64360_PCI_1_INTEGRATED_SRAM_BASE_ADDR_REMAP:
			enablebit = BIT15;
			break;
		}
	if (enablebit == 0xffff) return -1;
	devfn = PCI_DEVFN(0, pci_cfg_fcn);
	/* Internal registers do not have size register */
	if (enablebit != BIT9)
		MV_REG_WRITE(bar_size_reg, (bar_size - 1) & 0xfffff000);
	if (bar_size == 0) {
		if (pci_bus == 0) MV_SET_REG_BITS (MV64360_PCI_0_BASE_ADDR_REG_ENABLE, enablebit);
		if (pci_bus == 1) MV_SET_REG_BITS (MV64360_PCI_1_BASE_ADDR_REG_ENABLE, enablebit);
	} else {
		if (pci_bus == 0) MV_RESET_REG_BITS (MV64360_PCI_0_BASE_ADDR_REG_ENABLE, enablebit);
		if (pci_bus == 1) MV_RESET_REG_BITS (MV64360_PCI_1_BASE_ADDR_REG_ENABLE, enablebit);
	}

	if (remap_reg != 0)
		MV_REG_WRITE(remap_reg, cpu_base_addr);
	if (remap_reg_hi != 0)
		MV_REG_WRITE(remap_reg_hi, cpu_base_addr_hi);
	MV_REG_READ(remap_reg); /* Flush FIFO */

	save_exclude = mv64360_pci_exclude_bridge;
	mv64360_pci_exclude_bridge = FALSE;
	val = mv64360_hw_read_config_dword (pci_bus, hose->first_busno, PCI_SLOT(devfn),
					    PCI_FUNC(devfn), pci_cfg_hdr_offset, 0);
	val &= 0x0000000f;
	mv64360_hw_write_config_dword (pci_bus, hose->first_busno, PCI_SLOT(devfn),
				       PCI_FUNC(devfn), pci_cfg_hdr_offset, 0,
				       pci_base_addr | val);
	/* If 64bit BAR, Write high PCI address to next BAR PCI address */
	if ((val & 0x6) == 0x4)
		mv64360_hw_write_config_dword (pci_bus, hose->first_busno,
					       PCI_SLOT(devfn), PCI_FUNC(devfn),
					       pci_cfg_hdr_offset + 0x4, 0, pci_base_addr_hi);
	mv64360_pci_exclude_bridge = save_exclude;
	
	return 0;
} /* mv64360_pci_slave_set_window() */

int
mv64360_pci_slave_cs_set_window(struct pci_controller *hose,
				u32 window,
				u32 pci_base_addr,
				u32 pci_base_addr_hi,
				u32 cpu_base_addr,
				u32 cpu_base_addr_hi,
				u32 size)
{
	static u32
	pci_cs_windows[MV64360_PCI_BUSES][MV64360_PCI_CS_WINDOWS][5] = {
		{ /* PCI 0 */
			{ 0, 0x10,
			  MV64360_PCI_0_CS_0_BANK_SIZE,
			  MV64360_PCI_0_CS_0_BASE_ADDR_REMAP,
			  MV64360_PCI_0_CS_0_BASE_HIGH_ADDR_REMAP },
			{ 0, 0x18,
			  MV64360_PCI_0_CS_1_BANK_SIZE,
			  MV64360_PCI_0_CS_1_BASE_ADDR_REMAP,
			  MV64360_PCI_0_CS_1_BASE_HIGH_ADDR_REMAP },
			{ 1, 0x10,
			  MV64360_PCI_0_CS_2_BANK_SIZE,
			  MV64360_PCI_0_CS_2_BASE_ADDR_REMAP,
			  MV64360_PCI_0_CS_2_BASE_HIGH_ADDR_REMAP },
			{ 1, 0x18,
			  MV64360_PCI_0_CS_3_BANK_SIZE,
			  MV64360_PCI_0_CS_3_BASE_ADDR_REMAP,
			  MV64360_PCI_0_CS_3_BASE_HIGH_ADDR_REMAP },
		},
		{ /* PCI 1 */
			{ 0, 0x10,
			  MV64360_PCI_1_CS_0_BANK_SIZE,
			  MV64360_PCI_1_CS_0_BASE_ADDR_REMAP,
			  MV64360_PCI_1_CS_0_BASE_HIGH_ADDR_REMAP },
			{ 0, 0x18,
			  MV64360_PCI_1_CS_1_BANK_SIZE,
			  MV64360_PCI_1_CS_1_BASE_ADDR_REMAP,
			  MV64360_PCI_1_CS_1_BASE_HIGH_ADDR_REMAP },
			{ 1, 0x10,
			  MV64360_PCI_1_CS_2_BANK_SIZE,
			  MV64360_PCI_1_CS_2_BASE_ADDR_REMAP,
			  MV64360_PCI_1_CS_2_BASE_HIGH_ADDR_REMAP },
			{ 1, 0x18,
			  MV64360_PCI_1_CS_3_BANK_SIZE,
			  MV64360_PCI_1_CS_3_BASE_ADDR_REMAP,
			  MV64360_PCI_1_CS_3_BASE_HIGH_ADDR_REMAP }
		}
	}; /* pci_cs_windows[][][] */
	int	pci_bus;
	int	rc = -1;

	if (window < MV64360_PCI_CS_WINDOWS) {
		pci_bus = (hose->first_busno == 0) ? 0 : 1;

		rc = mv64360_pci_slave_set_window(
				hose,
				pci_base_addr,
				pci_base_addr_hi,
				cpu_base_addr,
				cpu_base_addr_hi,
				size,
				pci_cs_windows[pci_bus][window][0],
				pci_cs_windows[pci_bus][window][1],
				pci_cs_windows[pci_bus][window][2],
				pci_cs_windows[pci_bus][window][3],
				pci_cs_windows[pci_bus][window][4]);
	}

	return rc;
} /* mv64360_pci_slave_cs_set_window() */

int
mv64360_pci_slave_dev_cs_set_window(struct pci_controller *hose,
				    u32 window,
				    u32 pci_base_addr,
				    u32 pci_base_addr_hi,
				    u32 cpu_base_addr,
				    u32 size)
{
	static u32
	pci_cs_windows[MV64360_PCI_BUSES][MV64360_PCI_DEV_CS_WINDOWS][4] = {
		{ /* PCI 0 */
			{ 2, 0x10,
			  MV64360_PCI_0_DEVCS_0_BANK_SIZE,
			  MV64360_PCI_0_DEVCS_0_BASE_ADDR_REMAP },
			{ 2, 0x18,
			  MV64360_PCI_0_DEVCS_1_BANK_SIZE,
			  MV64360_PCI_0_DEVCS_1_BASE_ADDR_REMAP },
			{ 2, 0x20,
			  MV64360_PCI_0_DEVCS_2_BANK_SIZE,
			  MV64360_PCI_0_DEVCS_2_BASE_ADDR_REMAP },
			{ 3, 0x10,
			  MV64360_PCI_0_DEVCS_3_BANK_SIZE,
			  MV64360_PCI_0_DEVCS_3_BASE_ADDR_REMAP },
		},
		{ /* PCI 1 */
			{ 2, 0x10,
			  MV64360_PCI_1_DEVCS_0_BANK_SIZE,
			  MV64360_PCI_1_DEVCS_0_BASE_ADDR_REMAP },
			{ 2, 0x18,
			  MV64360_PCI_1_DEVCS_1_BANK_SIZE,
			  MV64360_PCI_1_DEVCS_1_BASE_ADDR_REMAP },
			{ 2, 0x20,
			  MV64360_PCI_1_DEVCS_2_BANK_SIZE,
			  MV64360_PCI_1_DEVCS_2_BASE_ADDR_REMAP },
			{ 3, 0x10,
			  MV64360_PCI_1_DEVCS_3_BANK_SIZE,
			  MV64360_PCI_1_DEVCS_3_BASE_ADDR_REMAP }
		}
	}; /* pci_cs_windows[][][] */
	int	pci_bus;
	int	rc = -1;

	if (window < MV64360_PCI_CS_WINDOWS) {
		pci_bus = (hose->first_busno == 0) ? 0 : 1;

		rc = mv64360_pci_slave_set_window(
				hose,
				pci_base_addr,
				pci_base_addr_hi,
				cpu_base_addr,
				0,/*Dev CS does not support 64bit addressing*/
				size,
				pci_cs_windows[pci_bus][window][0],
				pci_cs_windows[pci_bus][window][1],
				pci_cs_windows[pci_bus][window][2],
				pci_cs_windows[pci_bus][window][3],
				0);
	}

	return rc;
} /* mv64360_pci_slave_dev_cs_set_window() */

int
mv64360_pci_slave_boot_set_window(struct pci_controller *hose,
				  u32 pci_base_addr,
				  u32 pci_base_addr_hi,
				  u32 cpu_base_addr,
				  u32 size)
{
	int	rc;
	int	pci_bus;
	pci_bus = (hose->first_busno == 0) ? 0 : 1;
	if (pci_bus == 0)
	  rc = mv64360_pci_slave_set_window(hose,
					    pci_base_addr,
					    pci_base_addr_hi,
					    cpu_base_addr,
					    0,
					    size,
					    3,
					    0x18,
					    MV64360_PCI_0_DEVCS_BOOT_BANK_SIZE,
					    MV64360_PCI_0_DEVCS_BOOTCS_BASE_ADDR_REMAP,
					    0);
	else
	  rc = mv64360_pci_slave_set_window(hose,
					    pci_base_addr,
					    pci_base_addr_hi,
					    cpu_base_addr,
					    0,
					    size,
					    3,
					    0x18,
					    MV64360_PCI_1_DEVCS_BOOT_BANK_SIZE,
					    MV64360_PCI_1_DEVCS_BOOTCS_BASE_ADDR_REMAP,
					    0);

	return rc;
} /* mv64360_pci_slave_boot_set_window() */
int
mv64360_pci_slave_internal_sram_set_window(struct pci_controller *hose,
					   u32 pci_base_addr,
					   u32 pci_base_addr_hi,
					   u32 cpu_base_addr,
					   u32 size)
{
	int	rc;
	int	pci_bus;
	pci_bus = (hose->first_busno == 0) ? 0 : 1;

	if (pci_bus == 0)
		rc = mv64360_pci_slave_set_window(hose,
						  pci_base_addr,
						  pci_base_addr_hi,
						  cpu_base_addr,
						  0,
						  size,
						  1,
						  0x20,
						  MV64360_PCI_0_INTERNAL_SRAM_BAR_SIZE,
						  MV64360_PCI_0_INTEGRATED_SRAM_BASE_ADDR_REMAP,
						  0);
	else
		rc = mv64360_pci_slave_set_window(hose,
						  pci_base_addr,
						  pci_base_addr_hi,
						  cpu_base_addr,
						  0,
						  size,
						  1,
						  0x20,
						  MV64360_PCI_1_INTERNAL_SRAM_BAR_SIZE,
						  MV64360_PCI_1_INTEGRATED_SRAM_BASE_ADDR_REMAP,
						  0);
	
	return rc;
} /* mv64360_pci_slave_internal_sram_set_window() */

int
mv64360_pci_slave_internal_registers_set_window(struct pci_controller *hose,
						u32 pci_base_addr,
						u32 pci_base_addr_hi,
						u32 cpu_base_addr,
						u32 size)
{
	int	rc;
	int	pci_bus;
	pci_bus = (hose->first_busno == 0) ? 0 : 1;
	if (pci_bus == 0)
		rc = mv64360_pci_slave_set_window(hose,
						  pci_base_addr,
						  pci_base_addr_hi,
						  cpu_base_addr,
						  0,
						  size,
						  0,
						  0x20,
						  0,
						  0,
						  0);
	else
		rc = mv64360_pci_slave_set_window(hose,
						  pci_base_addr,
						  pci_base_addr_hi,
						  cpu_base_addr,
						  0,
						  size,
						  0,
						  0x20,
						  0,
						  0,
						  0);
	
	return rc;
} /* mv64360_pci_slave_internal_sram_set_window() */


int
mv64360_pci_slave_p2p_mem_set_window(struct pci_controller *hose,
				     u32 window,
				     u32 pci_base_addr,
				     u32 pci_base_addr_hi,
				     u32 other_bus_base_addr,
				     u32 other_bus_base_addr_hi,
				     u32 size)
{
	static u32
	pci_p2p_mem_windows[MV64360_PCI_BUSES][MV64360_PCI_P2P_MEM_WINDOWS][5]={
		{ /* PCI 0 */
			{ 4, 0x10,
			  MV64360_PCI_0_P2P_MEM0_BAR_SIZE,
			  MV64360_PCI_0_P2P_MEM0_BASE_ADDR_REMAP_LOW,
			  MV64360_PCI_0_P2P_MEM0_BASE_ADDR_REMAP_HIGH },
			{ 4, 0x18,
			  MV64360_PCI_0_P2P_MEM1_BAR_SIZE,
			  MV64360_PCI_0_P2P_MEM1_BASE_ADDR_REMAP_LOW,
			  MV64360_PCI_0_P2P_MEM1_BASE_ADDR_REMAP_HIGH },
		},
		{ /* PCI 1 */
			{ 4, 0x10,
			  MV64360_PCI_1_P2P_MEM0_BAR_SIZE,
			  MV64360_PCI_1_P2P_MEM0_BASE_ADDR_REMAP_LOW,
			  MV64360_PCI_1_P2P_MEM0_BASE_ADDR_REMAP_HIGH },
			{ 4, 0x18,
			  MV64360_PCI_1_P2P_MEM1_BAR_SIZE,
			  MV64360_PCI_1_P2P_MEM1_BASE_ADDR_REMAP_LOW,
			  MV64360_PCI_1_P2P_MEM1_BASE_ADDR_REMAP_HIGH },
		}
	}; /* pci_p2p_mem_windows[][][] */
	int	pci_bus;
	int	rc = -1;

	if (window < MV64360_PCI_P2P_MEM_WINDOWS) {
		pci_bus = (hose->first_busno == 0) ? 0 : 1;

		rc = mv64360_pci_slave_set_window(
				hose,
				pci_base_addr,
				pci_base_addr_hi,
				other_bus_base_addr,
				other_bus_base_addr_hi,
				size,
				pci_p2p_mem_windows[pci_bus][window][0],
				pci_p2p_mem_windows[pci_bus][window][1],
				pci_p2p_mem_windows[pci_bus][window][2],
				pci_p2p_mem_windows[pci_bus][window][3],
				pci_p2p_mem_windows[pci_bus][window][4]);
	}

	return rc;
} /* mv64360_pci_slave_p2p_mem_set_window() */

int
mv64360_pci_slave_p2p_io_set_window(struct pci_controller *hose,
				    u32 pci_base_addr,
				    u32 other_bus_base_addr,
				    u32 size)
{
	int	rc;
	int     pci_bus;

	pci_bus = (hose->first_busno == 0) ? 0 : 1;
	if (pci_bus == 0)
	  rc = mv64360_pci_slave_set_window(hose,
					    pci_base_addr,
					    0,
					    other_bus_base_addr,
					    0,
					    size,
					    4,
					    0x20,
					    MV64360_PCI_0_P2P_I_O_BAR_SIZE,
					    MV64360_PCI_0_P2P_I_O_BASE_ADDR_REMAP,
					    0);
	else
	  rc = mv64360_pci_slave_set_window(hose,
					    pci_base_addr,
					    0,
					    other_bus_base_addr,
					    0,
					    size,
					    4,
					    0x20,
					    MV64360_PCI_1_P2P_I_O_BAR_SIZE,
					    MV64360_PCI_1_P2P_I_O_BASE_ADDR_REMAP,
					    0);

	return rc;
} /* mv64360_pci_slave_p2p_io_set_window() */


void mv64360_pcibios_fixup(void) {
	struct pci_dev *dev;
	u16 status;
	u8 id, next, current, bus;
	u32 val;
	return;
	/*
	 * go over scanned PCI devices and modify their
	 * PCI-X capability list, if valid.
	 * (Code done by Mordi Blaunstein)
	 */

	/* First update MV64360 PCI-X capability */
	for (bus = 0; bus < num_hoses; bus++) {
		val = mv64360_hw_read_config_dword (bus, 0, 0, 0, 0x60, 0);
		if ((val & 0xff) == 0x07) {
			mv64360_hw_write_config_dword (bus, 0, 0, 0, 0x60, 0,
						       (0x0000 << 16) | (val & 0x0000ffff));
		}
	}
	pci_for_each_dev( dev )  {
		/* Update all devices cache line size to be 32 bytes ! */
		pci_write_config_byte (dev , 0xc , L1_CACHE_LINE_SIZE >> 2);
		
		/* Mordi - PCI-X capabiltiies List initialization */
		pci_read_config_word (dev, PCI_STATUS , &status );
		/* Check if capbilty pointer is located at offset PCI_CAPABILITY_LIST */
		if ( ( status & BIT4) != 0 ) {
			pci_read_config_byte (dev, PCI_CAPABILITY_LIST , &next);
			do {
				current = next;
				pci_read_config_byte (dev, current , &id);
				if ( id == 0x7 ) {
					pci_read_config_dword (dev, current, &val);
					pci_write_config_word (dev, (current + 0x2 ), 0x0000 );
				}
				else if ( id > 0x7 ) {
					printk(KERN_ERR "mv64360_pcibios_fixup: ERROR Capabilty ID %x on device # %x\n",id,dev->devfn);
				}
				pci_read_config_byte (dev, (current + 0x1 ),&next);
			} while ( next != 0x0 );
		}
	}
}

/*
 *****************************************************************************
 *
 *	PCI Control Configuration Routines
 *
 *****************************************************************************
 */

void
mv64360_cpu0_sync_barrier (int bus) {
	if (bus == 0) MV_REG_WRITE (MV64360_CPU_0_SYNC_BARRIER_TRIGGER,
				    BIT3 | BIT2 | BIT0);
	if (bus == 1) MV_REG_WRITE (MV64360_CPU_0_SYNC_BARRIER_TRIGGER,
				    BIT3 | BIT2 | BIT1);

	while (MV_REG_READ (MV64360_CPU_0_SYNC_BARRIER_VIRTUAL)) {};
}

int
mv64360_pci_acc_cntl_set_window(u32 pci_bus,
				u32 window,
				u32 base_addr_hi,
				u32 base_addr_lo,
				u32 size,
				u32 features)
{
	unsigned int new_size, temp, shift;
	static u32
	pci_acc_cntl_windows[MV64360_PCI_BUSES][MV64360_PCI_ACC_CNTL_WINDOWS][3] = {
		{ /* PCI 0 */
			{ MV64360_PCI_0_ACCESS_CONTROL_BASE_0_HIGH,
			  MV64360_PCI_0_ACCESS_CONTROL_BASE_0_LOW,
			  MV64360_PCI_0_ACCESS_CONTROL_SIZE_0 },

			{ MV64360_PCI_0_ACCESS_CONTROL_BASE_1_HIGH,
			  MV64360_PCI_0_ACCESS_CONTROL_BASE_1_LOW,
			  MV64360_PCI_0_ACCESS_CONTROL_SIZE_1 },

			{ MV64360_PCI_0_ACCESS_CONTROL_BASE_2_HIGH,
			  MV64360_PCI_0_ACCESS_CONTROL_BASE_2_LOW,
			  MV64360_PCI_0_ACCESS_CONTROL_SIZE_2 },

			{ MV64360_PCI_0_ACCESS_CONTROL_BASE_3_HIGH,
			  MV64360_PCI_0_ACCESS_CONTROL_BASE_3_LOW,
			  MV64360_PCI_0_ACCESS_CONTROL_SIZE_3 },

		},
		{ /* PCI 1 */
			{ MV64360_PCI_1_ACCESS_CONTROL_BASE_0_HIGH,
			  MV64360_PCI_1_ACCESS_CONTROL_BASE_0_LOW,
			  MV64360_PCI_1_ACCESS_CONTROL_SIZE_0 },

			{ MV64360_PCI_1_ACCESS_CONTROL_BASE_1_HIGH,
			  MV64360_PCI_1_ACCESS_CONTROL_BASE_1_LOW,
			  MV64360_PCI_1_ACCESS_CONTROL_SIZE_1 },

			{ MV64360_PCI_1_ACCESS_CONTROL_BASE_2_HIGH,
			  MV64360_PCI_1_ACCESS_CONTROL_BASE_2_LOW,
			  MV64360_PCI_1_ACCESS_CONTROL_SIZE_2 },

			{ MV64360_PCI_1_ACCESS_CONTROL_BASE_3_HIGH,
			  MV64360_PCI_1_ACCESS_CONTROL_BASE_3_LOW,
			  MV64360_PCI_1_ACCESS_CONTROL_SIZE_3 },
			
		},
		
	}; /* pci_acc_cntl_windows[][][] */
	int	rc = -1;
	
	if ((pci_bus < MV64360_PCI_BUSES) &&
	    (window < MV64360_PCI_ACC_CNTL_WINDOWS)) {
		
		if (size >= _64K) {
			new_size = size >> 12;
			/* Checking that the size is a sequence of '1' followed by a
			   sequence of '0' starting from LSB to MSB. */
			temp = new_size - 1;
			for(shift = 0; shift < 20; shift++) {
				temp = temp >> shift;
				if((temp & 0x1) == 0) {
					/* Either we got to the last '1' */
					/* or the size is not valid       */
					if(temp > 0x0)
						return -1;
					else
						break;
				}
			}
		} else {
			/* size < 64KByte */
			MV_REG_WRITE (pci_acc_cntl_windows[pci_bus][window][0], 0);
			MV_REG_WRITE (pci_acc_cntl_windows[pci_bus][window][1], 0);
			MV_REG_WRITE (pci_acc_cntl_windows[pci_bus][window][2], 0);
			return 0;
		}			
		MV_REG_WRITE (pci_acc_cntl_windows[pci_bus][window][0], base_addr_hi);
		MV_REG_WRITE (pci_acc_cntl_windows[pci_bus][window][1],
			      (base_addr_lo & 0xfffff000) | features);
		MV_REG_WRITE (pci_acc_cntl_windows[pci_bus][window][2], (size - (1 << 12)) & 0xfffff000);
	}
	
	return rc;
} /* mv64360_pci_acc_cntl_set_window() */


/*
 *****************************************************************************
 *
 *	MV64360's Register Base Address Routines
 *
 *****************************************************************************
 */

/*
 * mv64360_remap_bridge_regs()
 *
 * Move the bridge's register to the specified base address.
 * Assume that there are no other windows overlapping this area and that
 * all but the highest 3 nibbles are 0.
 */
int
mv64360_set_base(u32 new_base)
{
	u32	val;
	int	limit = 100000;
	int	rc = 0;

	MV_REG_WRITE(MV64360_INTERNAL_SPACE_BASE_ADDR, val);
	
	iounmap((void *)mv64360_base);
	mv64360_base = (u32)ioremap((new_base & 0xfffff000),
				    MV64360_INTERNAL_SPACE_SIZE);

	wmb();
	mb();

	do { /* Wait for bridge to move its regs */
		val = MV_REG_READ(MV64360_INTERNAL_SPACE_BASE_ADDR);
	} while ((val != 0xffffffff) && (limit-- > 0));

	if (limit <= 0) {
		rc = -1;
	}
	
	return rc;
} /* mv64360_remap_bridge_regs() */

/*
 * mv64360_get_base()
 *
 * Return the current virtual base address of the 64360's registers.
 */
int
mv64360_get_base(u32 *base)
{
	*base = mv64360_base;
	return 0;
} /* mv64360_remap_bridge_regs() */


int
mv64360_get_cpu_cs_base_and_size (int window, u32 *base, u32 *size) {
	u32 bitenable;
	static u32
	cpu_cs_windows[MV64360_CPU_CS_DECODE_WINDOWS][2] = {
		{ MV64360_CS_0_BASE_ADDR, MV64360_CS_0_SIZE },
		{ MV64360_CS_1_BASE_ADDR, MV64360_CS_1_SIZE },
		{ MV64360_CS_2_BASE_ADDR, MV64360_CS_2_SIZE },
		{ MV64360_CS_3_BASE_ADDR, MV64360_CS_3_SIZE },
	}; /* cpu_cs_windows[][] */

	bitenable = MV_REG_READ (MV64360_BASE_ADDR_ENABLE);

	if ((window >= 0) && (window <= MV64360_CPU_CS_DECODE_WINDOWS)) {
		if (bitenable & (1<< window)) {
			*base = 0;
			*size = 0;
		} else {
			*base = (MV_REG_READ (cpu_cs_windows[window][0]) << 16);
			*size = ((MV_REG_READ (cpu_cs_windows[window][1]) + 1) << 16);
		}
		return 0;
	}
	return -1;
}


/*
 *****************************************************************************
 *
 *	Exclude PCI config space access to bridge itself
 *
 *****************************************************************************
 */

/*
 * mv64360_exclude_pci_device()
 *
 * This routine causes the PCI subsystem to skip the PCI device in slot 0
 * (which is the MV64360 itself) unless explicitly allowed.
 */
int
mv64360_pci_exclude_device(u8 bus, u8 devfn)
{
	struct pci_controller	*hose;
	hose = pci_bus_to_hose(bus);
	/* Skip slot 0 on both hoses */
	if ((mv64360_pci_exclude_bridge == TRUE) &&
	    (PCI_SLOT(devfn) == 0) && (hose->first_busno == bus)) {
		return PCIBIOS_DEVICE_NOT_FOUND;
	}
	else {
		return PCIBIOS_SUCCESSFUL;
	}
} /* mv64360_pci_exclude_device() */

/*
 *****************************************************************************
 *
 *	Misc Routines
 *
 *****************************************************************************
 */
/*
 * mv64360_get_mem_size()
 *
 * Read memory controller's registers to determine the amount of memory
 * in the system.  Assumes that the memory controller has been set up
 * to the proper memory size.
 */
ulong __init
mv64360_get_mem_size(void)
{
	static u32
	cpu_cs_windows[MV64360_CPU_CS_DECODE_WINDOWS] = {
		MV64360_CS_0_SIZE,
		MV64360_CS_1_SIZE,
		MV64360_CS_2_SIZE,
		MV64360_CS_3_SIZE
	}; /* cpu_cs_windows[][] */
	u32	window, total, enable;
	
	total = 0;
	enable = MV_REG_READ (MV64360_BASE_ADDR_ENABLE);
	for (window=0; window<MV64360_CPU_CS_DECODE_WINDOWS; window++) {
		if (enable & (1 << window)) continue;
		total += ((MV_REG_READ(cpu_cs_windows[window]) + 1) << 16);
	}
	return total;
} /* mv64360_get_mem_size() */


/*
 * PCI configuration operations definition
 */

u32
mv64360_hw_read_config_dword (u8 local_bus, u8 dest_bus, u8 dev, u8 function, u8 offset, u8 type) {
	u32 cfg;
	u32 val;
	cfg = BIT31 | (dest_bus << 16) | (dev << 11) | (function << 8) | (offset & ~0x3);
	if (local_bus == 0) {
		u32 temp;
		/* Disable Master abort interrupt */
		temp = MV_REG_READ (MV64360_PCI_0_ERROR_MASK);
		MV_REG_WRITE (MV64360_PCI_0_ERROR_MASK, temp & ~BIT8);

		MV_REG_WRITE (MV64360_PCI_0_CONFIG_ADDR, cfg);
		wmb();
		val = MV_REG_READ (MV64360_PCI_0_CONFIG_DATA_VIRTUAL_REG);
		wmb();

		/* Reenable Master abort interrupt */
		MV_REG_WRITE (MV64360_PCI_0_ERROR_CAUSE, ~BIT8);
		MV_REG_WRITE (MV64360_PCI_0_ERROR_MASK, temp & ~BIT8);

	} else {
		u32 temp;
		/* Disable Master abort interrupt */
		temp = MV_REG_READ (MV64360_PCI_1_ERROR_MASK);
		MV_REG_WRITE (MV64360_PCI_1_ERROR_MASK, temp & ~BIT8);

		MV_REG_WRITE (MV64360_PCI_1_CONFIG_ADDR, cfg);
		wmb();
		val = MV_REG_READ (MV64360_PCI_1_CONFIG_DATA_VIRTUAL_REG);
		wmb();

		/* Reenable Master abort interrupt */
		MV_REG_WRITE (MV64360_PCI_1_ERROR_CAUSE, ~BIT8);
		MV_REG_WRITE (MV64360_PCI_1_ERROR_MASK, temp & ~BIT8);
	}
	return val;
}
		
void
mv64360_hw_write_config_dword (u8 local_bus, u8 dest_bus, u8 dev, u8 function, u8 offset, u8 type, u32 val) {
	u32 cfg;
	cfg = BIT31 | (dest_bus << 16) | (dev << 11) | (function << 8) | (offset & ~0x3);
	if (local_bus == 0) {
		MV_REG_WRITE (MV64360_PCI_0_CONFIG_ADDR, cfg);
		wmb ();
		MV_REG_WRITE (MV64360_PCI_0_CONFIG_DATA_VIRTUAL_REG, val);
		wmb();
	} else {
		MV_REG_WRITE (MV64360_PCI_1_CONFIG_ADDR, cfg);
		wmb ();
		MV_REG_WRITE (MV64360_PCI_1_CONFIG_DATA_VIRTUAL_REG, val);
		wmb();
	}
}
		
#if defined(CONFIG_SERIAL_TEXT_DEBUG)
/*
 *****************************************************************************
 *
 *	Low-level MPSC/UART I/O routines
 *
 *****************************************************************************
 */

/* Rabeeh - TBD - the following functions need rework */
#if 1
/*
 * mv64360_putc()
 *
 * Dump a character out the MPSC port for mv64360_mpsc_progress
 * this assumes the baud rate has already been set up and the
 * MPSC initialized by the bootloader or firmware.
 */

static inline void
gt_putc(char c){
	u32	data;
	u32 	dummy;

	dummy = MV_REG_READ(MV64360_MPSC_CAUSE_REG(0));
	MV_REG_WRITE(MV64360_MPSC_CHANNEL_REG1(0), c);
	mb();
	data = MV_REG_READ(MV64360_MPSC_CHANNEL_REG2(0));
	data |= MV64360_MPSC_UART_CR_TTCS;
	MV_REG_WRITE(MV64360_MPSC_CHANNEL_REG2(0), data);
	mb();

	while (MV_REG_READ(MV64360_MPSC_CHANNEL_REG2(0))
				& MV64360_MPSC_UART_CR_TTCS) ;
}

void
puthex(unsigned long val){

	int i;

	for (i = 7;  i >= 0;  i--) {
		gt_putc("0123456789ABCDEF"[(val>>28) & 0x0f]);
		val <<= 4;
	}
	gt_putc('\r');
	gt_putc('\n');	
	
}


void
mv64360_mpsc_progress(char *s, unsigned short hex){
	/* spit stuff out the MV64360 mpsc */
	
	volatile char	c;
	while ((c = *s++) != 0){
		gt_putc(c);
		if ( c == '\n' ) gt_putc('\r');	
	}
	gt_putc('\n');
	gt_putc('\r');

	return;
}
#endif

#endif /* CONFIG_DEBUG_TEXT */
