/*
 * arch/ppc/kernel/mv64360_pic.c
 * 
 * Interrupt controller support for Marvell's MV64360.
 *
 * Author: Rabeeh Khoury <rabeeh@galileo.co.il>
 * Based on MV64360 PIC written by
 * Chris Zankel <chris@mvista.com>
 * Mark A. Greer <mgreer@mvista.com>
 *
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

/*
 * This file contains the specific functions to support the MV64360
 * interrupt controller.
 *
 * The MV64360 has two main interrupt registers (high and low) that
 * summarizes the interrupts generated by the units of the MV64360.
 * Each bit is assigned to an interrupt number, where the low register
 * are assigned from IRQ0 to IRQ31 and the high cause register
 * from IRQ32 to IRQ63 
 * The GPP (General Purpose Pins) interrupts are assigned from IRQ64 (GPP0) 
 * to IRQ95 (GPP31). 
 * get_irq() returns the lowest interrupt number that is currently asserted.
 *
 * Note: 
 *  - This driver does not initialize the GPP when used as an interrupt 
 *    input.
 */

#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/stddef.h>
#include <linux/delay.h>
#include <linux/irq.h>

#include <asm/io.h>
#include <asm/processor.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/mv64360.h>

#define MV64460_VEN_DEV_ID (PCI_DEVICE_ID_GALILEO_MV64460 << 16 | \
			PCI_VENDOR_ID_GALILEO)

#ifdef CONFIG_IRQ_ALL_CPUS
#error "The mv64360 does not support yet distribution of IRQs on all CPUs"
#endif
/* ========================== forward declaration ========================== */

static void mv64360_unmask_irq(unsigned int);
static void mv64360_mask_irq(unsigned int);
static void mv64360_cpu_error_int_handler(int, void *, struct pt_regs *);
static void mv64360_sram_error_int_handler(int, void *, struct pt_regs *);
static void mv64360_pci_error_int_handler(int, void *, struct pt_regs *);

/* ========================== local declarations =========================== */

struct hw_interrupt_type mv64360_pic = {
	" MV64360_PIC ",	/* typename */
	NULL,			/* startup */
	NULL,			/* shutdown */
	mv64360_unmask_irq,	/* enable */
	mv64360_mask_irq,	/* disable */
	mv64360_mask_irq,	/* ack */
	NULL,			/* end */
	NULL			/* set_affinity */
};

extern u32 mv64360_revision; /* Revision of the chip */

char cpu_intr[] = "MV64360 CPU interface error";
char sram_intr[] = "MV64360 internal sram error";
char pci0_intr[] = "MV64360 PCI 0 error";
char pci1_intr[] = "MV64360 PCI 1 error";

u32 mv64360_irq_base = 0;	/* MV64360 handles the next 96 IRQs from here */

/* mv64360_init_irq()
 *
 *  This function initializes the interrupt controller. It assigns
 *  all interrupts from IRQ0 to IRQ95 to the mv64360 interrupt controller.
 *
 * Input Variable(s):
 *  None.
 *
 * Outpu. Variable(s):
 *  None.
 *
 * Returns:
 *  void
 *
 * Note:
 *  We register all GPP inputs as interrupt source, but disable them.
 */

__init void
mv64360_init_irq(void)
{
	u32 i;

	if (ppc_md.progress)
		ppc_md.progress("mv64360_init_irq: enter", 0x0);

	ppc_cached_irq_mask[0] = 0;
	ppc_cached_irq_mask[1] = 0x0f000000;	/* Enable GPP intrs */
	ppc_cached_irq_mask[2] = 0;

	/* disable all interrupts and clear current interrupts */
	MV_REG_WRITE(MV64360_GPP_INTERRUPT_CAUSE, 0);
	MV_REG_WRITE(MV64360_GPP_INTERRUPT_MASK0, ppc_cached_irq_mask[2]);
	MV_REG_WRITE(MV64360_CPU_INTERRUPT0_MASK_LOW, ppc_cached_irq_mask[0]);
	MV_REG_WRITE(MV64360_CPU_INTERRUPT0_MASK_HIGH, ppc_cached_irq_mask[1]);

	/* use the mv64360 for all (possible) interrupt sources */
	for (i = mv64360_irq_base; i < (mv64360_irq_base + 96); i++) {
		/* All interrupts are level interrupts */
		irq_desc[i].status |= IRQ_LEVEL;
		irq_desc[i].handler = &mv64360_pic;
	}

	/* Register CPU interface error interrupt handler */
	request_irq(3, mv64360_cpu_error_int_handler,
		SA_INTERRUPT, cpu_intr, 0);
	MV_REG_WRITE(MV64360_CPU_ERROR_MASK, 0x000000ff);

	/* Register internal SRAM error interrupt handler */
	request_irq(13, mv64360_sram_error_int_handler,
		SA_INTERRUPT, sram_intr, 0);

	/*
	 * Handle MV64460 erratum FEr PCI-#11 (PCI internal data parity
	 * error set incorrectly)
	 */ 
	i = mv64360_hw_read_config_dword(0, 0, 0, 0, PCI_VENDOR_ID, 0);
	if (i == MV64460_VEN_DEV_ID) {
		i = 0x00a50c25;
		if (mv64360_revision <= 1)
			i &= ~BIT0;
	}

	/* Register PCI 0 error interrupt handler */
	request_irq(12, mv64360_pci_error_int_handler,
		SA_INTERRUPT, pci0_intr, (void *) 0);
	MV_REG_WRITE(MV64360_PCI_0_ERROR_MASK, i);

	/* Register PCI 1 error interrupt handler */
	request_irq(16, mv64360_pci_error_int_handler,
		SA_INTERRUPT, pci1_intr, (void *) 1);
	MV_REG_WRITE(MV64360_PCI_1_ERROR_MASK, i);

	if (ppc_md.progress)
		ppc_md.progress("mv64360_init_irq: exit", 0x0);
}


/* mv64360_get_irq()
 *
 *  This function returns the lowest interrupt number of all interrupts that
 *  are currently asserted. 
 *
 * Input Variable(s):
 *  struct pt_regs*	not used
 *
 * Output Variable(s):
 *  None.
 *
 * Returns:
 *  int	<interrupt number> or -2 (bogus interrupt)
 * 
 */
int
mv64360_get_irq(struct pt_regs *regs)
{
	int irq;
	int irq_gpp;

#ifdef CONFIG_SMP
	/* 
	 * Second CPU gets only doorbell (message) interrupts.
	 * The doorbell interrupt is BIT28 in the main interrupt low cause reg.
	 */
	int cpu_nr = smp_processor_id();
	if (cpu_nr == 1) {
		irq = MV_REG_READ(MV64360_MAIN_INTERRUPT_CAUSE_LOW);
		if (!(irq & BIT28))
			return -1;
		return 28;
	}
#endif

	irq = MV_REG_READ(MV64360_MAIN_INTERRUPT_CAUSE_LOW);
	irq = __ilog2((irq & 0x3dfffffe) & ppc_cached_irq_mask[0]);
	if (irq == -1) {
		irq = MV_REG_READ(MV64360_MAIN_INTERRUPT_CAUSE_HIGH);
		irq = __ilog2((irq & 0x1f0003f7) & ppc_cached_irq_mask[1]);
		if (irq == -1) {
			irq = -2;	/* bogus interrupt, should never happen */
		} else {
			if ((irq >= 24) && (irq < 28)) {
				irq_gpp =
					MV_REG_READ
					(MV64360_GPP_INTERRUPT_CAUSE);
				irq_gpp =
					__ilog2(irq_gpp &
					ppc_cached_irq_mask[2]);

				if (irq_gpp == -1) {
					irq = -2;
				} else {
					irq = irq_gpp + 64;
					MV_REG_WRITE
						(MV64360_GPP_INTERRUPT_CAUSE,
						~(1 << (irq - 64)));
				}
			} else {
				irq += 32;
			}
		}
	}

	if (irq < 0) {
		return (irq);
	} else {
		return (mv64360_irq_base + irq);
	}
}

/* mv64360_unmask_irq()
 *
 *  This function enables an interrupt.
 *
 * Input Variable(s):
 *  unsigned int	interrupt number (IRQ0...IRQ95).
 *
 * Output Variable(s):
 *  None.
 *
 * Returns:
 *  void
 */

static void
mv64360_unmask_irq(unsigned int irq)
{
#ifdef CONFIG_SMP
	int cpu_nr = smp_processor_id();
	/* second CPU gets only doorbell interrupts */
	if ((irq - mv64360_irq_base) == 28) {
		MV_SET_REG_BITS(MV64360_CPU_INTERRUPT1_MASK_LOW, BIT28);
		return;
	}
#endif
	irq -= mv64360_irq_base;
	if (irq > 31) {
		if (irq > 63) {
			/* unmask GPP irq */
			MV_REG_WRITE(MV64360_GPP_INTERRUPT_MASK0,
				ppc_cached_irq_mask[2] |= (1 << (irq - 64)));
		} else {
			/* mask high interrupt register */
			MV_REG_WRITE(MV64360_CPU_INTERRUPT0_MASK_HIGH,
				ppc_cached_irq_mask[1] |= (1 << (irq - 32)));
		}
	} else {
		/* mask low interrupt register */
		MV_REG_WRITE(MV64360_CPU_INTERRUPT0_MASK_LOW,
			ppc_cached_irq_mask[0] |= (1 << irq));
	}
	/*
	 * Need to drain the mv64360 pipeline to ensure that the irq has
	 * been masked before returning.  Reading any mv64360 register
	 * is sufficient.
	 */
	MV_REG_READ(MV64360_GPP_INTERRUPT_MASK0);
}


/* mv64360_mask_irq()
 *
 *  This function disables the requested interrupt.
 *
 * Input Variable(s):
 *  unsigned int	interrupt number (IRQ0...IRQ95).
 *
 * Output Variable(s):
 *  None.
 *
 * Returns:
 *  void
 */

static void
mv64360_mask_irq(unsigned int irq)
{
#ifdef CONFIG_SMP
	int cpu_nr = smp_processor_id();
	if ((irq - mv64360_irq_base) == 28) {
		MV_RESET_REG_BITS(MV64360_CPU_INTERRUPT1_MASK_LOW, BIT28);
		return;
	}
#endif
	irq -= mv64360_irq_base;
	if (irq > 31) {
		if (irq > 63) {
			/* mask GPP irq */
			MV_REG_WRITE(MV64360_GPP_INTERRUPT_MASK0,
				ppc_cached_irq_mask[2] &= ~(1 << (irq - 64)));
		} else {
			/* mask high interrupt register */
			MV_REG_WRITE(MV64360_CPU_INTERRUPT0_MASK_HIGH,
				ppc_cached_irq_mask[1] &= ~(1 << (irq - 32)));
		}
	} else {
		/* mask low interrupt register */
		MV_REG_WRITE(MV64360_CPU_INTERRUPT0_MASK_LOW,
			ppc_cached_irq_mask[0] &= ~(1 << irq));
	}

}

static void
mv64360_cpu_error_int_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 val;
	val = MV_REG_READ(MV64360_CPU_ERROR_CAUSE);
	printk(KERN_ERR
		"mv64360_cpu_error_int_handler: Error on CPU interface - Cause regiser 0x%08x\n",
		val);
	printk(KERN_ERR "\tCPU error register dump:\n");
	printk(KERN_ERR "\tAddress low  0x%08x\n",
		MV_REG_READ(MV64360_CPU_ERROR_ADDR_LOW));
	printk(KERN_ERR "\tAddress high 0x%08x\n",
		MV_REG_READ(MV64360_CPU_ERROR_ADDR_HIGH));
	printk(KERN_ERR "\tData low     0x%08x\n",
		MV_REG_READ(MV64360_CPU_ERROR_DATA_LOW));
	printk(KERN_ERR "\tData high    0x%08x\n",
		MV_REG_READ(MV64360_CPU_ERROR_DATA_HIGH));
	printk(KERN_ERR "\tParity       0x%08x\n",
		MV_REG_READ(MV64360_CPU_ERROR_PARITY));
	MV_REG_WRITE(MV64360_CPU_ERROR_CAUSE, 0);
}

static void
mv64360_sram_error_int_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	printk(KERN_ERR
		"mv64360_sram_error_int_handler: Error in internal SRAM - Cause register 0x%08x\n",
		MV_REG_READ(MV64360_SRAM_ERROR_CAUSE));
	printk(KERN_ERR "\tSRAM error register dump:\n");
	printk(KERN_ERR "\tAddress Low  0x%08x\n",
		MV_REG_READ(MV64360_SRAM_ERROR_ADDR));
	printk(KERN_ERR "\tAddress High 0x%08x\n",
		MV_REG_READ(MV64360_SRAM_ERROR_ADDR_HIGH));
	printk(KERN_ERR "\tData Low     0x%08x\n",
		MV_REG_READ(MV64360_SRAM_ERROR_DATA_LOW));
	printk(KERN_ERR "\tData High    0x%08x\n",
		MV_REG_READ(MV64360_SRAM_ERROR_DATA_HIGH));
	printk(KERN_ERR "\tParity       0x%08x\n",
		MV_REG_READ(MV64360_SRAM_ERROR_DATA_PARITY));
	MV_REG_WRITE(MV64360_SRAM_ERROR_CAUSE, 0);
}

static void
mv64360_pci_error_int_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 val;
	unsigned int pci_bus = (unsigned int) dev_id;
	if (pci_bus == 0) {	/* Error on PCI 0 */
		val = MV_REG_READ(MV64360_PCI_0_ERROR_CAUSE);
		printk(KERN_ERR
			"mv64360_pci_error_int_handler: Error in PCI %d Interface\n",
			pci_bus);
		printk(KERN_ERR "\tPCI %d error register dump:\n", pci_bus);
		printk(KERN_ERR "\tCause register 0x%08x\n", val);
		printk(KERN_ERR "\tAddress Low    0x%08x\n",
			MV_REG_READ(MV64360_PCI_0_ERROR_ADDR_LOW));
		printk(KERN_ERR "\tAddress High   0x%08x\n",
			MV_REG_READ(MV64360_PCI_0_ERROR_ADDR_HIGH));
		printk(KERN_ERR "\tAttribute      0x%08x\n",
			MV_REG_READ(MV64360_PCI_0_ERROR_ATTRIBUTE));
		printk(KERN_ERR "\tCommand        0x%08x\n",
			MV_REG_READ(MV64360_PCI_0_ERROR_COMMAND));
		MV_REG_WRITE(MV64360_PCI_0_ERROR_CAUSE, ~val);
	}
	if (pci_bus == 1) {	/* Error on PCI 1 */
		val = MV_REG_READ(MV64360_PCI_1_ERROR_CAUSE);
		printk(KERN_ERR
			"mv64360_pci_error_int_handler: Error in PCI %d Interface\n",
			pci_bus);
		printk(KERN_ERR "\tPCI %d error register dump:\n", pci_bus);
		printk(KERN_ERR "\tCause register 0x%08x\n", val);
		printk(KERN_ERR "\tAddress Low    0x%08x\n",
			MV_REG_READ(MV64360_PCI_1_ERROR_ADDR_LOW));
		printk(KERN_ERR "\tAddress High   0x%08x\n",
			MV_REG_READ(MV64360_PCI_1_ERROR_ADDR_HIGH));
		printk(KERN_ERR "\tAttribute      0x%08x\n",
			MV_REG_READ(MV64360_PCI_1_ERROR_ATTRIBUTE));
		printk(KERN_ERR "\tCommand        0x%08x\n",
			MV_REG_READ(MV64360_PCI_1_ERROR_COMMAND));
		MV_REG_WRITE(MV64360_PCI_1_ERROR_CAUSE, ~val);
	}
}
