/*
 * (C) Copyright 2004
 * John W. Linville, LVL7 Systems, Inc., linville@lvl7.com.
 *
 * Modifications for single 16-bit flash (and non-MPC8xx).
 *
 * (C) Copyright 2000-2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <mpc8xx.h>

#ifndef CFG_ENV_ADDR
#define CFG_ENV_ADDR    (CFG_FLASH_BASE + CFG_ENV_OFFSET)
#endif

flash_info_t    flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips    */

/*-----------------------------------------------------------------------
 * Functions
 */
static ulong flash_get_size (vu_short *addr, flash_info_t *info);
static int write_word (flash_info_t *info, ulong dest, ushort data);

/*-----------------------------------------------------------------------
 */

unsigned long flash_init (void)
{
    unsigned long size_b0;
    int i;

    /* Init: no FLASHes known */
    for (i=0; i<CFG_MAX_FLASH_BANKS; ++i) {
        flash_info[i].flash_id = FLASH_UNKNOWN;
    }

    /* Static FLASH Bank configuration here - FIXME XXX */

    size_b0 = flash_get_size((vu_short *)CFG_FLASH_BASE0_PRELIM, &flash_info[0]);

    if (flash_info[0].flash_id == FLASH_UNKNOWN) {
        printf ("## Unknown FLASH on Bank 0 - Size = 0x%08lx = %ld MB\n",
            size_b0, size_b0<<20);
    }

#if CFG_MONITOR_BASE >= CFG_FLASH_BASE
    /* monitor protection ON by default */
    flash_protect(FLAG_PROTECT_SET,
              CFG_MONITOR_BASE,
              CFG_MONITOR_BASE+CFG_MONITOR_LEN-1,
              &flash_info[0]);
#endif

#ifdef    CFG_ENV_IS_IN_FLASH
    /* ENV protection ON by default */
    flash_protect(FLAG_PROTECT_SET,
              CFG_ENV_ADDR,
              CFG_ENV_ADDR+CFG_ENV_SIZE-1,
              &flash_info[0]);
#endif

    flash_info[0].size = size_b0;

    return (size_b0);
}

/*-----------------------------------------------------------------------
 */
void flash_print_info(flash_info_t *info)
{
    int i;

    if (info->flash_id == FLASH_UNKNOWN) {
        printf ("missing or unknown FLASH type\n");
        return;
    }

    switch (info->flash_id & FLASH_VENDMASK) {
        case FLASH_MAN_AMD:
            printf ("AMD ");
            break;
        case FLASH_MAN_FUJ:
            printf ("FUJITSU ");
            break;
        case FLASH_MAN_MX:
            printf ("MXIC ");
            break;
        case FLASH_MAN_INTEL:
            printf ("Intel ");
            break;
        default:
            printf ("Unknown Vendor ");
            break;
    }

    switch (info->flash_id & FLASH_TYPEMASK) {
        case FLASH_AM640B:
            printf ("AM29LV640B (64 Mbit, bottom boot sect)\n");
            break;
        case FLASH_AM640T:
            printf ("AM29LV640T (64 Mbit, top boot sector)\n");
            break;
        case FLASH_MX640B:
            printf ("MX29LV640B (64 Mbit, bottom boot sect)\n");
            break;
        case FLASH_MX640T:
            printf ("MX29LV640T (64 Mbit, top boot sector)\n");
            break;
        case FLASH_28F640J3A:
            printf ("28F640J3A (64 Mbit, uniform sector)\n");
            break;
        default:
            printf ("Unknown Chip Type: 0x%x\n", info->flash_id);
            break;
    }

    printf ("  Size: %ld MB in %d Sectors\n",
        info->size >> 20, info->sector_count);

    printf ("  Sector Start Addresses:");
    for (i=0; i<info->sector_count; ++i) {
        if ((i % 5) == 0)
            printf ("\n   ");
        printf (" %08lX%s",
            info->start[i],
            info->protect[i] ? " (RO)" : "     "
        );
    }
    printf ("\n");
    return;
}

/*
 * The following code cannot be run from FLASH!
 */

static ulong flash_get_size (vu_short *addr, flash_info_t *info)
{
    short i;
    ushort value;
    ulong base = (ulong)addr;

    /* Write auto select command: read Manufacturer ID */
    addr[0x0555 ^ 0x01] = 0x00AA;
    addr[0x02AA ^ 0x01] = 0x0055;
    addr[0x0555 ^ 0x01] = 0x0090;

    value = addr[0x00 ^ 0x01];

    switch ((ulong)((value << 16) | value)) {
        case AMD_MANUFACT:
            info->flash_id = FLASH_MAN_AMD;
            break;
        case FUJ_MANUFACT:
            info->flash_id = FLASH_MAN_FUJ;
            break;
        case MX_MANUFACT:
            info->flash_id = FLASH_MAN_MX;
            break;
        case INTEL_MANUFACT:
            info->flash_id = FLASH_MAN_INTEL;
            break;
        default:
            printf("Unknown Flash Manufacturing ID: %x\n", ((value << 16) | value));
            info->flash_id = FLASH_UNKNOWN;
            info->sector_count = 0;
            info->size = 0;
            return (0);            /* no or unknown flash    */
    }

    value = addr[0x01 ^ 0x01];              /* device ID (part 1)    */
    switch(info->flash_id) {
        case FLASH_MAN_AMD:
        case FLASH_MAN_FUJ:
            if (value != 0x227e)
                return (0);

            value = addr[0x0e ^ 0x01];      /* device ID (part 2)    */
            if (value != 0x2210)
                return (0);

            value = addr[0x0f ^ 0x01];      /* device ID (part 3)    */
            break;
        case FLASH_MAN_MX:
        case FLASH_MAN_INTEL:
            break;

        default: /* shouldn't get here... */
            return (0);
    }

    /* set up sector start address table */
    switch ((ulong)((value << 16) | value)) {
        case AMD_ID_LV640T:
            info->flash_id = FLASH_AM640T;
            info->sector_count = 135;
            info->size = 0x00800000;
            for (i = 0; i < info->sector_count; i++) {
                info->start[i] = base;
                /*
                * The last 8 sectors are 8 kB,
                * all the other ones  are 64 kB
                */
                base += (i < (info->sector_count - 8))
                    ?  1 * (64 << 10)
                    :  1 * ( 8 << 10);
            }
            break;                /* => 8 MB        */

        case AMD_ID_LV640B:
            info->flash_id = FLASH_AM640B;
            info->sector_count = 135;
            info->size = 0x00800000;
            for (i = 0; i < info->sector_count; i++) {
                info->start[i] = base;
                /*
                 * The first 8 sectors are 8 kB,
                 * all the other ones  are 64 kB
                 */
                base += (i < 8)
                    ?  1 * ( 8 << 10)
                    :  1 * (64 << 10);
            }
            break;                /* => 8 MB        */

        case MX_ID_LV640T:
            info->flash_id = FLASH_MX640T;
            info->sector_count = 135;
            info->size = 0x00800000;
            for (i = 0; i < info->sector_count; i++) {
                info->start[i] = base;
                /*
                 * The last 8 sectors are 8 kB,
                 * all the other ones  are 64 kB
                 */
                base += (i < (info->sector_count - 8))
                    ?  1 * (64 << 10)
                    :  1 * ( 8 << 10);
            }
            break;                /* => 8 MB        */

        case MX_ID_LV640B:
            info->flash_id = FLASH_MX640B;
            info->sector_count = 135;
            info->size = 0x00800000;
            for (i = 0; i < info->sector_count; i++) {
                info->start[i] = base;
                /*
                 * The first 8 sectors are 8 kB,
                 * all the other ones  are 64 kB
                 */
                base += (i < 8)
                    ?  1 * ( 8 << 10)
                    :  1 * (64 << 10);
            }
            break;                /* => 8 MB        */

        case INTEL_ID_28F640J3A:
            info->flash_id = FLASH_28F640J3A;
            info->sector_count = 64;        /* 64 128k sectors */
            info->size = 0x00800000;
            for (i = 0; i < info->sector_count; i++) {
                info->start[i] = base;
                base += (128 * 1024);
            }
            break;                /* => 8 MB        */

        default:
            printf("Unknown Device ID: %x\n", ((value << 16) | value));
            info->flash_id = FLASH_UNKNOWN;
            return (0);            /* => no or unknown flash */
    }

    /* check for protected sectors */
    for (i = 0; i < info->sector_count; i++) {
        /* read sector protection at sector address, (A7 .. A0) = 0x02 */
        /* D0 = 1 if protected */
        info->protect[i] = addr[0x02 ^ 0x01] & 1;
    }

    *addr = 0x00F0;    /* reset bank */
    return (info->size);
}

/*-----------------------------------------------------------------------
 */

int flash_erase (flash_info_t *info, int s_first, int s_last)
{
    vu_short *addr = (vu_short*)(info->start[0]);
    int flag, prot, sect, l_sect = 0;
    ulong start, now, last;

    if ((s_first < 0) || (s_first > s_last)) {
        if (info->flash_id == FLASH_UNKNOWN) {
            printf ("- missing\n");
        } else {
            printf ("- no sectors to erase\n");
        }
        return 1;
    }

    if (info->flash_id == FLASH_UNKNOWN) {
        printf ("Can't erase unknown flash type %08lx - aborted\n",
            info->flash_id);
        return 1;
    }

    prot = 0;
    for (sect=s_first; sect<=s_last; ++sect) {
        if (info->protect[sect]) {
            prot++;
        }
    }

    if (prot) {
        printf ("- Warning: %d protected sectors will not be erased!\n", prot);
    } else {
        printf ("\n");
    }

    /* Disable interrupts which might cause a timeout here */
    flag = disable_interrupts();

    if (info->flash_id != FLASH_28F640J3A) {
        addr[0x0555 ^ 0x01] = 0x00AA;
        addr[0x02AA ^ 0x01] = 0x0055;
        addr[0x0555 ^ 0x01] = 0x0080;
        addr[0x0555 ^ 0x01] = 0x00AA;
        addr[0x02AA ^ 0x01] = 0x0055;
    
        /* Start erase on unprotected sectors */
        for (sect = s_first; sect<=s_last; sect++) {
            if (info->protect[sect] == 0) {    /* not protected */
                addr = (vu_short*)(info->start[sect]);
                addr[0] = 0x0030;
                l_sect = sect;
            }
        }

        /* re-enable interrupts if necessary */
        if (flag)
            enable_interrupts();

        /* wait at least 80us - let's wait 1 ms */
        udelay (1000);

        /*
         * We wait for the last triggered sector
         */
        if (l_sect) {
            start = get_timer(0);
            last  = start;
            addr = (vu_short*)(info->start[l_sect]);
            while ((addr[0] & 0x0080) != 0x0080) {
                if ((now = get_timer(start)) > CFG_FLASH_ERASE_TOUT) {
                    printf ("Timeout\n");
                    return 1;
                }
                /* show that we're waiting */
                if ((now - last) > CFG_HZ) {    /* every second */
                    putc ('.');
                    last = now;
                }
            }
        } 

        /* reset to read mode */
        addr = (unsigned short *)info->start[0];
        addr[0] = 0x00F0;    /* reset bank */
    }
    else {  /* Intel Flash Erase */
        /* Start erase on unprotected sectors */
        for (sect = s_first; sect<=s_last; sect++) {
            if (info->protect[sect] == 0) {    /* not protected */
                addr = (vu_short*)(info->start[sect]);
                addr[0] = 0x0020;       /* Erase Sector */
                addr[0] = 0x00d0;       /* Sector erase confirm */

                udelay (1000);

                start = get_timer(0);
                last  = start;
                while ((addr[0] & 0x0080) != 0x0080) {
                    if ((now = get_timer(start)) > CFG_FLASH_ERASE_TOUT) {
                        printf ("Timeout\n");
                        return 1;
                    }
                    /* show that we're waiting */
                    if ((now - last) > CFG_HZ) {    /* every second */
                        putc ('.');
                        last = now;
                    }
                }
            }
        }

        /* re-enable interrupts if necessary */
        if (flag)
            enable_interrupts();

        /* reset to read mode */
        addr = (unsigned short *)info->start[0];
        addr[0] = 0x00FF;    /* reset bank */
    }

    printf (" done\n");
    return 0;
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */

int write_buff(flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
    ulong cp, wp;
    ushort data;
    int i, l, rc;

    wp = (addr & ~1);    /* get lower word aligned address */

    /*
     * handle unaligned start bytes
     */
    if ((l = addr - wp) != 0) {
        data = 0;
        for (i=0, cp=wp; i<l; ++i, ++cp) {
            data = (data << 8) | (*(uchar *)cp);
        }
        for (; i<2 && cnt>0; ++i) {
            data = (data << 8) | *src++;
            --cnt;
            ++cp;
        }
        for (; cnt==0 && i<2; ++i, ++cp) {
            data = (data << 8) | (*(uchar *)cp);
        }

        if ((rc = write_word(info, wp, data)) != 0) {
            return (rc);
        }
        wp += 2;
    }

    /*
     * handle word aligned part
     */
    while (cnt >= 2) {
        data = 0;
        for (i=0; i<2; ++i) {
            data = (data << 8) | *src++;
        }
        if ((rc = write_word(info, wp, data)) != 0) {
            return (rc);
        }
        wp  += 2;
        cnt -= 2;
    }

    if (cnt == 0) {
        return (0);
    }

    /*
     * handle unaligned tail bytes
     */
    data = 0;
    for (i=0, cp=wp; i<2 && cnt>0; ++i, ++cp) {
        data = (data << 8) | *src++;
        --cnt;
    }
    for (; i<2; ++i, ++cp) {
        data = (data << 8) | (*(uchar *)cp);
    }

    return (write_word(info, wp, data));
}

/*-----------------------------------------------------------------------
 * Write a word to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */
static int write_word(flash_info_t *info, ulong dest, ushort data)
{
    vu_short *addr = (vu_short*)(info->start[0]);
    ulong start;
    int flag;

    /* Check if Flash is (sufficiently) erased */
    if ((*((vu_short *)dest) & data) != data) {
        return (2);
    }

    /* Disable interrupts which might cause a timeout here */
    flag = disable_interrupts();

    if (info->flash_id != FLASH_28F640J3A) {
        addr[0x0555 ^ 0x01] = 0x00AA;
        addr[0x02AA ^ 0x01] = 0x0055;
        addr[0x0555 ^ 0x01] = 0x00A0;
    }
    else {
        *((vu_short *)dest) = 0x0040;
    }

    *((vu_short *)dest) = data;

    /* re-enable interrupts if necessary */
    if (flag)
        enable_interrupts();

    /* data polling for D7 */
    start = get_timer(0);
    while ((*((vu_short *)dest) & 0x0080) != 0x0080) {
        if (get_timer(start) > CFG_FLASH_WRITE_TOUT) {
            *addr = 0x00FF;
            return (1);
        }
    }
    
    if (info->flash_id == FLASH_28F640J3A) {
        *addr = 0x00FF;
    }

    return (0);
}
